# Copyright (c) 2013 Chris Lucas, <chris@chrisjlucas.com>
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import os.path
import time
import urllib.error
import urllib.parse
import urllib.request
import xmlrpc.client
from xmlrpc.client import ServerProxy

import rtorrent.file
import rtorrent.peer
import rtorrent.rpc
import rtorrent.tracker
from rtorrent.common import convert_version_tuple_to_str, is_valid_port
from rtorrent.group import Group
from rtorrent.lib.torrentparser import TorrentParser
from rtorrent.lib.xmlrpc.requests_transport import RequestsTransport
from rtorrent.lib.xmlrpc.scgi import SCGIServerProxy
from rtorrent.rpc import Method
from rtorrent.torrent import Torrent

__version__ = "0.2.9"
__author__ = "Chris Lucas"
__maintainer__ = "Dustyn Gibson"
__contact__ = "miigotu@gmail.com"
__license__ = "MIT"

MIN_RTORRENT_VERSION = (0, 9, 0)
MIN_RTORRENT_VERSION_STR = convert_version_tuple_to_str(MIN_RTORRENT_VERSION)
MAX_RETRIES = 5


class RTorrent:

    """ Create a new rTorrent connection """
    rpc_prefix = None

    def __init__(self, uri, username=None, password=None,
                 verify=False, sp=None, sp_kwargs=None, tp_kwargs=None):
        self.uri = uri  # : From X{__init__(self, url)}

        self.username = username
        self.password = password

        self.schema = urllib.parse.urlsplit(uri).scheme

        if sp:
            self.sp = sp
        elif self.schema in ['http', 'https']:
            self.sp = ServerProxy
            if self.schema == 'https':
                self.isHttps = True
            else:
                self.isHttps = False
        elif self.schema == 'scgi':
            self.sp = SCGIServerProxy
        else:
            raise NotImplementedError()

        self.sp_kwargs = sp_kwargs or {}

        self.tp_kwargs = tp_kwargs or {}

        self.torrents = []  # : List of L{Torrent} instances
        self._rpc_methods = []  # : List of rTorrent RPC methods
        self._torrent_cache = []
        self._client_version_tuple = ()

        if verify is True:
            self._verify_conn()

    def _get_conn(self):
        """Get ServerProxy instance"""
        if self.username is not None and self.password is not None:
            if self.schema == 'scgi':
                raise NotImplementedError()

            if 'authtype' not in self.tp_kwargs:
                authtype = None
            else:
                authtype = self.tp_kwargs['authtype']

            if 'check_ssl_cert' not in self.tp_kwargs:
                check_ssl_cert = True
            else:
                check_ssl_cert = self.tp_kwargs['check_ssl_cert']

            if 'proxies' not in self.tp_kwargs:
                proxies = None
            else:
                proxies = self.tp_kwargs['proxies']

            return self.sp(
                self.uri,
                transport=RequestsTransport(
                    use_https=self.isHttps,
                    authtype=authtype,
                    username=self.username,
                    password=self.password,
                    check_ssl_cert=check_ssl_cert,
                    proxies=proxies),
                **self.sp_kwargs
            )

        return self.sp(self.uri, **self.sp_kwargs)

    def _verify_conn(self):
        # check for rpc methods that should be available
        assert "system.client_version" in self._get_rpc_methods(
        ), "Required RPC method not available."
        assert "system.library_version" in self._get_rpc_methods(
        ), "Required RPC method not available."

        # minimum rTorrent version check
        assert self._meets_version_requirement() is True,\
            "Error: Minimum rTorrent version required is {0}".format(
            MIN_RTORRENT_VERSION_STR)

    def _meets_version_requirement(self):
        return self._get_client_version_tuple() >= MIN_RTORRENT_VERSION

    def _get_client_version_tuple(self):
        conn = self._get_conn()

        if not self._client_version_tuple:
            if not hasattr(self, "client_version"):
                setattr(self, "client_version",
                        conn.system.client_version())

            rtver = getattr(self, "client_version")
            self._client_version_tuple = tuple([int(i) for i in
                                                rtver.split(".")])

        return self._client_version_tuple

    def _update_rpc_methods(self):
        self._rpc_methods = self._get_conn().system.listMethods()

        return self._rpc_methods

    def _get_rpc_methods(self):
        """ Get list of raw RPC commands

        @return: raw RPC commands
        @rtype: list
        """

        return self._rpc_methods or self._update_rpc_methods()

    def get_torrents(self, view="main"):
        """Get list of all torrents in specified view

        @return: list of L{Torrent} instances

        @rtype: list

        @todo: add validity check for specified view
        """
        self.torrents = []
        methods = torrent.methods
        retriever_methods = [m for m in methods
                             if m.is_retriever() and m.is_available(self)]

        m = rpc.Multicall(self)
        m.add("d.multicall2", '', view, "d.hash=",
              *[method.rpc_call + "=" for method in retriever_methods])

        results = m.call()[0]  # only sent one call, only need first result

        for result in results:
            results_dict = {}
            # build results_dict
            # result[0] is the info_hash
            for m, r in zip(retriever_methods, result[1:]):
                results_dict[m.varname] = rpc.process_result(m, r)

            self.torrents.append(
                Torrent(self, info_hash=result[0], **results_dict)
            )

        self._manage_torrent_cache()
        return self.torrents

    def _manage_torrent_cache(self):
        """Carry tracker/peer/file lists over to new torrent list"""
        for torrent in self._torrent_cache:
            new_torrent = common.find_torrent(torrent.info_hash,
                                                       self.torrents)
            if new_torrent is not None:
                new_torrent.files = torrent.files
                new_torrent.peers = torrent.peers
                new_torrent.trackers = torrent.trackers

        self._torrent_cache = self.torrents

    def _get_load_function(self, file_type, start, verbose):
        """Determine correct "load torrent" RPC method"""
        func_name = None
        if file_type == "url":
            # url strings can be input directly
            if start and verbose:
                func_name = "load.start_verbose"
            elif start:
                func_name = "load.start"
            elif verbose:
                func_name = "load.verbose"
            else:
                func_name = "load.normal"
        elif file_type in ["file", "raw"]:
            if start and verbose:
                func_name = "load.raw_start_verbose"
            elif start:
                func_name = "load.raw_start"
            elif verbose:
                func_name = "load.raw_verbose"
            else:
                func_name = "load.raw"

        return func_name

    def load_magnet(self, magneturl, info_hash, **kwargs):
        start = kwargs.pop('start', False)
        verbose = kwargs.pop('verbose', False)
        verify_load = kwargs.pop('verify_load', True)
        max_retries = kwargs.pop('max_retries', 5)
        params = kwargs.pop('params', [])
        target = kwargs.pop('target', '')

        p = self._get_conn()

        info_hash = info_hash.upper()

        func_name = self._get_load_function("url", start, verbose)

        # load magnet
        getattr(p, func_name)(target, magneturl, *params)

        if verify_load:
            i = 0
            while i < max_retries:
                for torrent in self.get_torrents():
                    if torrent.info_hash == info_hash:
                        return True
                time.sleep(1)
                i += 1
            return False

        return True

    def load_torrent(self, torrent, **kwargs):
        """
        Load torrent into rTorrent (with various enhancements).
        @param torrent: can be a url, a path to a local file, or the raw data
        of a torrent file
        @type torrent: str
        @param start: start torrent when loaded
        @type start: bool
        @param verbose: print error messages to rTorrent log
        @type verbose: bool
        @param verify_load: verify that torrent was added to rTorrent successfully
        @type verify_load: bool
        @return: Depends on verify_load:
                 - if verify_load is True, (and the torrent was
                 loaded successfully), it'll return a L{Torrent} instance
                 - if verify_load is False, it'll return None
        @rtype: L{Torrent} instance or None
        @raise AssertionError: If the torrent wasn't successfully added to rTorrent
                               - Check L{TorrentParser} for the AssertionError's
                               it raises
        @note: Because this function includes url verification (if a url was input)
        as well as verification as to whether the torrent was successfully added,
        this function doesn't execute instantaneously. If that's what you're
        looking for, use load_torrent_simple() instead.
        """
        start = kwargs.pop('start', False)
        verbose = kwargs.pop('verbose', False)
        verify_load = kwargs.pop('verify_load', True)
        max_retries = kwargs.pop('max_retries', 5)
        params = kwargs.pop('params', [])
        target = kwargs.pop('target', '')

        p = self._get_conn()
        tp = TorrentParser(torrent)
        torrent = xmlrpc.client.Binary(tp._raw_torrent)
        info_hash = tp.info_hash

        func_name = self._get_load_function('raw', start, verbose)

        getattr(p, func_name)(target, torrent, *params)

        if verify_load:
            i = 0
            while i < max_retries:
                for torrent in self.get_torrents():
                    if torrent.info_hash == info_hash:
                        return True
                time.sleep(1)
                i += 1
            return False

        return True

    def load_torrent_simple(self, torrent, file_type, **kwargs):
        """Load torrent into rTorrent.
        @param torrent: can be a url, a path to a local file, or the raw data
        of a torrent file
        @type torrent: str
        @param file_type: valid options: 'url', 'file', or 'raw'
        @type file_type: str
        @param start: start torrent when loaded
        @type start: bool
        @param verbose: print error messages to rTorrent log
        @type verbose: bool
        @return: None
        @raise AssertionError: if incorrect file_type is specified
        @note: This function was written for speed, it includes no enhancements.
        If you input a url, it won't check if it's valid. You also can't get
        verification that the torrent was successfully added to rTorrent.
        Use load_torrent() if you would like these features.
        """
        start = kwargs.pop('start', False)
        verbose = kwargs.pop('verbose', False)
        params = kwargs.pop('params', [])
        target = kwargs.pop('target', '')

        p = self._get_conn()

        if file_type not in ['raw', 'file', 'url']:
            print("Invalid file_type, options are: 'url', 'file', 'raw'.")
            return False

        func_name = self._get_load_function(file_type, start, verbose)

        if file_type == 'file':
            # since we have to assume we're connected to a remote rTorrent
            # client, we have to read the file and send it to rT as raw
            if not os.path.isfile(torrent):
                print('Invalid path: {0}'.format(torrent))
                return False

            torrent = open(torrent, 'rb').read()

        finput = None
        if file_type in ['raw', 'file']:
            finput = xmlrpc.client.Binary(torrent)
        elif file_type in ['url', 'magnet']:
            finput = torrent

        getattr(p, func_name)(target, finput, *params)

        return True

    def get_views(self):
        p = self._get_conn()
        return p.view_list()

    def create_group(self, name, persistent=True, view=None):
        p = self._get_conn()

        if persistent is True:
            p.group.insert_persistent_view('', name)
        elif view is not None:
            p.group.insert('', name, view)

        self._update_rpc_methods()

    def get_group(self, name):
        assert name is not None, "group name required"

        _group = Group(self, name)
        _group.update()
        return _group

    def set_dht_port(self, port):
        """Set DHT port

        @param port: port
        @type port: int

        @raise AssertionError: if invalid port is given
        """
        assert is_valid_port(port), "Valid port range is 0-65535"
        self.dht_port = self._p.set_dht_port(port)

    def enable_check_hash(self):
        """Alias for set_check_hash(True)"""
        self.set_check_hash(True)

    def disable_check_hash(self):
        """Alias for set_check_hash(False)"""
        self.set_check_hash(False)

    def find_torrent(self, info_hash):
        """Frontend for rtorrent.common.find_torrent"""
        return common.find_torrent(info_hash, self.get_torrents())

    def poll(self):
        """ poll rTorrent to get latest torrent/peer/tracker/file information

        @note: This essentially refreshes every aspect of the rTorrent
        connection, so it can be very slow if working with a remote
        connection that has a lot of torrents loaded.

        @return: None
        """
        self.update()
        torrents = self.get_torrents()
        for t in torrents:
            t.poll()

    def update(self):
        """Refresh rTorrent client info

        @note: All fields are stored as attributes to self.

        @return: None
        """
        multicall = rpc.Multicall(self)
        retriever_methods = [m for m in methods
                             if m.is_retriever() and m.is_available(self)]
        for method in retriever_methods:
            multicall.add(method)

        multicall.call()


def _build_class_methods(class_obj):
    # multicall add class
    caller = lambda self, multicall, method, *args:\
        multicall.add(method, self.rpc_id, *args)

    caller.__doc__ = """Same as Multicall.add(), but with automatic inclusion
                        of the rpc_id

                        @param multicall: A L{Multicall} instance
                        @type: multicall: Multicall

                        @param method: L{Method} instance or raw rpc method
                        @type: Method or str

                        @param args: optional arguments to pass
                        """
    setattr(class_obj, "multicall_add", caller)


def __compare_rpc_methods(rt_new, rt_old):
    from pprint import pprint
    rt_new_methods = set(rt_new._get_rpc_methods())
    rt_old_methods = set(rt_old._get_rpc_methods())
    print("New Methods:")
    pprint(rt_new_methods - rt_old_methods)
    print("Methods not in new rTorrent:")
    pprint(rt_old_methods - rt_new_methods)


def __check_supported_methods(rt):
    from pprint import pprint
    supported_methods = set([m.rpc_call for m in
                             methods +
                             rtorrent.file.methods +
                             rtorrent.torrent.methods +
                             rtorrent.tracker.methods +
                             rtorrent.peer.methods])
    all_methods = set(rt._get_rpc_methods())

    print('Methods NOT in supported methods')
    pprint(all_methods - supported_methods)
    print('Supported methods NOT in all methods')
    pprint(supported_methods - all_methods)


methods = [
    # RETRIEVERS
    Method(RTorrent, 'get_xmlrpc_size_limit', 'network.xmlrpc.size_limit'),
    Method(RTorrent, 'get_proxy_address', 'network.proxy_address'),
    Method(RTorrent, 'get_file_split_suffix', 'system.file.split_suffix'),
    Method(RTorrent, 'get_global_up_limit', 'throttle.global_up.max_rate'),
    Method(RTorrent, 'get_max_memory_usage', 'pieces.memory.max'),
    Method(RTorrent, 'get_max_open_files', 'network.max_open_files'),
    Method(RTorrent, 'get_min_peers_seed', 'throttle.min_peers.seed'),
    Method(RTorrent, 'get_use_udp_trackers', 'trackers.use_udp'),
    Method(RTorrent, 'get_preload_min_size', 'pieces.preload.min_size'),
    Method(RTorrent, 'get_max_uploads', 'throttle.max_uploads'),
    Method(RTorrent, 'get_max_peers', 'throttle.max_peers.normal'),
    Method(RTorrent, 'get_timeout_sync', 'pieces.sync.timeout'),
    Method(RTorrent, 'get_receive_buffer_size', 'network.receive_buffer.size'),
    Method(RTorrent, 'get_split_file_size', 'system.file.split_size'),
    Method(RTorrent, 'get_dht_throttle', 'dht.throttle.name'),
    Method(RTorrent, 'get_max_peers_seed', 'throttle.max_peers.seed'),
    Method(RTorrent, 'get_min_peers', 'throttle.min_peers.normal'),
    Method(RTorrent, 'get_tracker_numwant', 'trackers.numwant'),
    Method(RTorrent, 'get_max_open_sockets', 'network.max_open_sockets'),
    Method(RTorrent, 'get_session_path', 'session.path'),
    Method(RTorrent, 'get_local_address', 'network.local_address'),
    Method(RTorrent, 'get_scgi_dont_route', 'network.scgi.dont_route'),
    Method(RTorrent, 'get_http_cacert', 'network.http.cacert'),
    Method(RTorrent, 'get_dht_port', 'dht.port'),
    Method(RTorrent, 'get_preload_type', 'pieces.preload.type'),
    Method(RTorrent, 'get_http_max_open', 'network.http.max_open'),
    Method(RTorrent, 'get_http_capath', 'network.http.capath'),
    Method(RTorrent, 'get_max_downloads_global', 'throttle.max_downloads.global'),
    Method(RTorrent, 'get_session_name', 'session.name'),
    Method(RTorrent, 'get_session_on_completion', 'session.on_completion'),
    Method(RTorrent, 'get_down_limit', 'throttle.global_down.max_rate'),
    Method(RTorrent, 'get_down_total', 'throttle.global_down.total'),
    Method(RTorrent, 'get_up_rate', 'throttle.global_up.rate'),
    Method(RTorrent, 'get_peer_exchange', 'protocol.pex'),
    Method(RTorrent, 'get_down_rate', 'throttle.global_down.rate'),
    Method(RTorrent, 'get_connection_seed', 'protocol.connection.seed'),
    Method(RTorrent, 'get_http_proxy_address', 'network.http.proxy_address'),
    Method(RTorrent, 'get_stats_preloaded', 'pieces.stats_preloaded'),
    Method(RTorrent, 'get_timeout_safe_sync', 'pieces.sync.timeout_safe'),
    Method(RTorrent, 'get_port_random', 'network.port_random'),
    Method(RTorrent, 'get_directory', 'directory.default'),
    Method(RTorrent, 'get_port_open', 'network.port_open'),
    Method(RTorrent, 'get_max_file_size', 'system.file.max_size'),
    Method(RTorrent, 'get_stats_not_preloaded', 'pieces.stats_not_preloaded'),
    Method(RTorrent, 'get_memory_usage', 'pieces.memory.current'),
    Method(RTorrent, 'get_connection_leech', 'protocol.connection.leech'),
    Method(RTorrent, 'get_hash_on_completion', 'pieces.hash.on_completion',
           boolean=True,
           ),
    Method(RTorrent, 'get_session_lock', 'session.use_lock'),
    Method(RTorrent, 'get_preload_min_rate', 'pieces.preload.min_rate'),
    Method(RTorrent, 'get_max_uploads_global', 'throttle.max_uploads.global'),
    Method(RTorrent, 'get_send_buffer_size', 'network.send_buffer.size'),
    Method(RTorrent, 'get_port_range', 'network.port_range'),
    Method(RTorrent, 'get_max_downloads_div', 'throttle.max_downloads.div'),
    Method(RTorrent, 'get_max_uploads_div', 'throttle.max_uploads.div'),
    Method(RTorrent, 'get_always_safe_sync', 'pieces.sync.always_safe'),
    Method(RTorrent, 'get_bind_address', 'network.bind_address'),
    Method(RTorrent, 'get_up_total', 'throttle.global_up.total'),
    Method(RTorrent, 'get_client_version', 'system.client_version'),
    Method(RTorrent, 'get_library_version', 'system.library_version'),
    Method(RTorrent, 'get_api_version', 'system.api_version',
           min_version=(0, 9, 1)
           ),
    Method(RTorrent, 'get_system_time', 'system.time',
           docstring="""Get the current time of the system rTorrent is running on
           @return: time (posix)
           @rtype: int""",
           ),

    # MODIFIERS
    Method(RTorrent, 'set_http_proxy_address', 'network.http.proxy_address.set'),
    Method(RTorrent, 'set_max_memory_usage', 'pieces.memory.max.set'),
    Method(RTorrent, 'set_max_file_size', 'system.file.max_size.set'),
    Method(RTorrent, 'set_bind_address', 'network.bind_address.set',
           docstring="""Set address bind
           @param arg: ip address
           @type arg: str
           """,
           ),
    Method(RTorrent, 'set_up_limit', 'throttle.global_up.max_rate.set',
           docstring="""Set global upload limit (in bytes)
           @param arg: speed limit
           @type arg: int
           """,
           ),
    Method(RTorrent, 'set_port_random', 'network.port_random.set'),
    Method(RTorrent, 'set_connection_leech', 'protocol.connection.leech.set'),
    Method(RTorrent, 'set_tracker_numwant', 'trackers.numwant.set'),
    Method(RTorrent, 'set_max_peers', 'throttle.max_peers.normal.set'),
    Method(RTorrent, 'set_min_peers', 'throttle.min_peers.normal.set'),
    Method(RTorrent, 'set_max_uploads_div', 'throttle.max_uploads.div.set'),
    Method(RTorrent, 'set_max_open_files', 'network.max_open_files.set'),
    Method(RTorrent, 'set_max_downloads_global', 'throttle.max_downloads.global.set'),
    Method(RTorrent, 'set_session_lock', 'session.use_lock.set'),
    Method(RTorrent, 'set_session_path', 'session.path.set'),
    Method(RTorrent, 'set_file_split_suffix', 'system.file.split_suffix.set'),
    Method(RTorrent, 'set_port_range', 'network.port_range.set'),
    Method(RTorrent, 'set_min_peers_seed', 'throttle.min_peers.seed.set'),
    Method(RTorrent, 'set_scgi_dont_route', 'network.scgi.dont_route.set'),
    Method(RTorrent, 'set_preload_min_size', 'pieces.preload.min_size.set'),
    Method(RTorrent, 'set_max_uploads_global', 'throttle.max_uploads.global.set'),
    Method(RTorrent, 'set_down_limit', 'throttle.global_down.max_rate.set',
           docstring="""Set global download limit (in bytes)
           @param arg: speed limit
           @type arg: int
           """,
           ),
    Method(RTorrent, 'set_preload_min_rate', 'pieces.preload.min_rate.set'),
    Method(RTorrent, 'set_max_peers_seed', 'throttle.max_peers.seed.set'),
    Method(RTorrent, 'set_max_uploads', 'throttle.max_uploads.set'),
    Method(RTorrent, 'set_session_on_completion', 'session.on_completion.set'),
    Method(RTorrent, 'set_max_open_http', 'network.http.max_open.set'),
    Method(RTorrent, 'set_directory', 'directory.default.set'),
    Method(RTorrent, 'set_http_cacert', 'network.http.cacert.set'),
    Method(RTorrent, 'set_dht_throttle', 'dht.throttle.name.set'),
    Method(RTorrent, 'set_proxy_address', 'network.proxy_address.set'),
    Method(RTorrent, 'set_split_file_size', 'system.file.split_size.set'),
    Method(RTorrent, 'set_receive_buffer_size', 'network.receive_buffer.size.set'),
    Method(RTorrent, 'set_use_udp_trackers', 'trackers.use_udp.set'),
    Method(RTorrent, 'set_connection_seed', 'protocol.connection.seed.set'),
    Method(RTorrent, 'set_xmlrpc_size_limit', 'network.xmlrpc.size_limit.set'),
    Method(RTorrent, 'set_xmlrpc_dialect', 'network.xmlrpc.dialect.set'),
    Method(RTorrent, 'set_always_safe_sync', 'pieces.sync.always_safe.set'),
    Method(RTorrent, 'set_http_capath', 'network.http.capath.set'),
    Method(RTorrent, 'set_send_buffer_size', 'network.send_buffer.size.set'),
    Method(RTorrent, 'set_max_downloads_div', 'throttle.max_downloads.div.set'),
    Method(RTorrent, 'set_session_name', 'session.name.set'),
    Method(RTorrent, 'set_port_open', 'network.port_open.set'),
    Method(RTorrent, 'set_timeout_sync', 'pieces.sync.timeout.set'),
    Method(RTorrent, 'set_peer_exchange', 'protocol.pex.set'),
    Method(RTorrent, 'set_local_address', 'network.local_address.set',
           docstring="""Set IP
           @param arg: ip address
           @type arg: str
           """,
           ),
    Method(RTorrent, 'set_timeout_safe_sync', 'pieces.sync.timeout_safe.set'),
    Method(RTorrent, 'set_preload_type', 'pieces.preload.type.set'),
    Method(RTorrent, 'set_hash_on_completion', 'pieces.hash.on_completion.set',
           docstring="""Enable/Disable hash checking on finished torrents
            @param arg: True to enable, False to disable
            @type arg: bool
            """,
           boolean=True,
           ),
]

_all_methods_list = [methods,
                     rtorrent.file.methods,
                     rtorrent.torrent.methods,
                     rtorrent.tracker.methods,
                     rtorrent.peer.methods,
                     ]

class_methods_pair = {
    RTorrent: methods,
    rtorrent.file.File: rtorrent.file.methods,
    torrent.Torrent: rtorrent.torrent.methods,
    rtorrent.tracker.Tracker: rtorrent.tracker.methods,
    rtorrent.peer.Peer: rtorrent.peer.methods,
}
for key, value in class_methods_pair.items():
    rpc._build_rpc_methods(key, value)
    _build_class_methods(key)



# -*- coding: utf-8 -*-

############################ Copyrights and license ############################
#                                                                              #
# Copyright 2012 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2012 Zearin <zearin@gonk.net>                                      #
# Copyright 2013 AKFish <akfish@gmail.com>                                     #
# Copyright 2013 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2014 Vincent Jacques <vincent@vincent-jacques.net>                 #
# Copyright 2016 Jannis Gebauer <ja.geb@me.com>                                #
# Copyright 2016 Peter Buckley <dx-pbuckley@users.noreply.github.com>          #
# Copyright 2018 Wan Liuyang <tsfdye@gmail.com>                                #
# Copyright 2018 sfdye <tsfdye@gmail.com>                                      #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

import github.GithubObject


class GitTreeElement(github.GithubObject.NonCompletableGithubObject):
    """
    This class represents GitTreeElements
    """

    def __repr__(self):
        return self.get__repr__({"sha": self._sha.value, "path": self._path.value})

    @property
    def mode(self):
        """
        :type: string
        """
        return self._mode.value

    @property
    def path(self):
        """
        :type: string
        """
        return self._path.value

    @property
    def sha(self):
        """
        :type: string
        """
        return self._sha.value

    @property
    def size(self):
        """
        :type: integer
        """
        return self._size.value

    @property
    def type(self):
        """
        :type: string
        """
        return self._type.value

    @property
    def url(self):
        """
        :type: string
        """
        return self._url.value

    def _initAttributes(self):
        self._mode = github.GithubObject.NotSet
        self._path = github.GithubObject.NotSet
        self._sha = github.GithubObject.NotSet
        self._size = github.GithubObject.NotSet
        self._type = github.GithubObject.NotSet
        self._url = github.GithubObject.NotSet

    def _useAttributes(self, attributes):
        if "mode" in attributes:  # pragma no branch
            self._mode = self._makeStringAttribute(attributes["mode"])
        if "path" in attributes:  # pragma no branch
            self._path = self._makeStringAttribute(attributes["path"])
        if "sha" in attributes:  # pragma no branch
            self._sha = self._makeStringAttribute(attributes["sha"])
        if "size" in attributes:  # pragma no branch
            self._size = self._makeIntAttribute(attributes["size"])
        if "type" in attributes:  # pragma no branch
            self._type = self._makeStringAttribute(attributes["type"])
        if "url" in attributes:  # pragma no branch
            self._url = self._makeStringAttribute(attributes["url"])



import random
from threading import Lock
from typing import TYPE_CHECKING

from .init_helpers import setup_gettext, setup_lib_path, sickchill_dir

setup_lib_path()

import rarfile

from sickchill.oldbeard.common import SD
from sickchill.oldbeard.numdict import NumDict

if TYPE_CHECKING:
    from .movies import MovieList

setup_gettext()

dynamic_strings = (
    _("Drama"),
    _("Mystery"),
    _("Science-Fiction"),
    _("Crime"),
    _("Action"),
    _("Comedy"),
    _("Thriller"),
    _("Animation"),
    _("Family"),
    _("Fantasy"),
    _("Adventure"),
    _("Horror"),
    _("Film-Noir"),
    _("Sci-Fi"),
    _("Romance"),
    _("Sport"),
    _("War"),
    _("Biography"),
    _("History"),
    _("Music"),
    _("Western"),
    _("News"),
    _("Sitcom"),
    _("Reality-TV"),
    _("Documentary"),
    _("Game-Show"),
    _("Musical"),
    _("Talk-Show"),
    _("Science-Fiction"),
)

__INITIALIZED__ = {}
ADBA_CONNECTION = None
ADD_SHOWS_WITH_YEAR = False
ADD_SHOWS_WO_DIR = False
ADDIC7ED_PASS = None
ADDIC7ED_USER = None
AIRDATE_EPISODES = False
ALLOW_HIGH_PRIORITY = False
ALLOWED_EXTENSIONS = "srt,nfo,srr,sfv"
ANIDB_PASSWORD = None
ANIDB_USE_MYLIST = False
ANIDB_USERNAME = None
ANIME_DEFAULT = False
ANIME_SPLIT_HOME = False
ANIME_SPLIT_HOME_IN_TABS = False
ANIMESUPPORT = False
ANON_REDIRECT = None
API_KEY = None
API_ROOT = None
AUTO_UPDATE = False
AUTOPOSTPROCESSOR_FREQUENCY = None
autoPostProcessorScheduler = None
BACKLOG_DAYS = 7
BACKLOG_FREQUENCY = None
BACKLOG_MISSING_ONLY = False
backlogSearchScheduler = None
BOXCAR2_ACCESSTOKEN = None
BOXCAR2_NOTIFY_ONDOWNLOAD = False
BOXCAR2_NOTIFY_ONSNATCH = False
BOXCAR2_NOTIFY_ONSUBTITLEDOWNLOAD = False
BRANCH = ""
BSDTAR_TOOL = rarfile.BSDTAR_TOOL
CACHE_DIR = None
CALENDAR_ICONS = False
CALENDAR_UNPROTECTED = False
CF_AUTH_DOMAIN = ""
CF_POLICY_AUD = ""
CFG = None
CHECK_PROPERS_INTERVAL = None
CLIENT_WEB_URLS = {"torrent": "", "newznab": ""}
COMING_EPS_DISPLAY_PAUSED = False
COMING_EPS_DISPLAY_SNATCHED = False
COMING_EPS_LAYOUT = None
COMING_EPS_MISSED_RANGE = None
COMING_EPS_SORT = None
CONFIG_FILE = ""
CONFIG_VERSION = 8
CPU_PRESET = None
CREATE_MISSING_SHOW_DIRS = False
CREATEPID = False
CUR_COMMIT_BRANCH = ""
CUR_COMMIT_HASH = None
CUSTOM_CSS = None
CUSTOM_CSS_PATH = None
DAEMON = None
DAILYSEARCH_FREQUENCY = 40
dailySearchScheduler = None
DATA_DIR = ""
DATE_PRESET = None
DBDEBUG = False
DEBUG = False
DEFAULT_AUTOPOSTPROCESSOR_FREQUENCY = 10
DEFAULT_BACKLOG_FREQUENCY = 21
DEFAULT_DAILYSEARCH_FREQUENCY = 40
DEFAULT_PAGE = "home"
DEFAULT_SHOWUPDATE_HOUR = random.randint(2, 4)
DEFAULT_UPDATE_FREQUENCY = 1
DELETE_FAILED = False
DELETE_NON_ASSOCIATED_FILES = False
DELRARCONTENTS = False
DEVELOPER = False
DISCORD_AVATAR_URL = "https://raw.githubusercontent.com/SickChill/SickChill/master/sickchill/gui/slick/images/sickchill-sc.png"
DISCORD_NAME = "SickChill"
DISCORD_NOTIFY_DOWNLOAD = None
DISCORD_NOTIFY_SNATCH = None
DISCORD_NOTIFY_SUBTITLEDOWNLOAD = None
DISCORD_TTS = False
DISCORD_WEBHOOK = None
DISPLAY_ALL_SEASONS = True
DISPLAY_SHOW_SPECIALS = False
DOWNLOAD_PROPERS = False
DOWNLOAD_URL = None
EMAIL_FROM = None
EMAIL_HOST = None
EMAIL_LIST = None
EMAIL_NOTIFY_ONDOWNLOAD = False
EMAIL_NOTIFY_ONPOSTPROCESS = False
EMAIL_NOTIFY_ONSNATCH = False
EMAIL_NOTIFY_ONSUBTITLEDOWNLOAD = False
EMAIL_PASSWORD = None
EMAIL_PORT = 25
EMAIL_SUBJECT = None
EMAIL_TLS = False
EMAIL_USER = None
EMBEDDED_SUBTITLES_ALL = False
EMBY_APIKEY = None
EMBY_HOST = None
ENABLE_HTTPS = False
ENCRYPTION_SECRET = None
ENCRYPTION_VERSION = 0
ENDED_SHOWS_UPDATE_INTERVAL = 7
EP_DEFAULT_DELETED_STATUS = None
events = None
EXTRA_SCRIPTS = []
FANART_API_KEY = "9b3afaf26f6241bdb57d6cc6bd798da7"
FANART_BACKGROUND = None
FANART_BACKGROUND_OPACITY = None
FILE_TIMESTAMP_TIMEZONE = None
FREEMOBILE_APIKEY = ""
FREEMOBILE_ID = ""
FREEMOBILE_NOTIFY_ONDOWNLOAD = False
FREEMOBILE_NOTIFY_ONSNATCH = False
FREEMOBILE_NOTIFY_ONSUBTITLEDOWNLOAD = False
FUZZY_DATING = False
gh = None
GIT_ORG = "SickChill"
GIT_PATH = None
GIT_REMOTE = ""
GIT_REMOTE_URL = ""
GIT_REPO = "SickChill"
GIT_RESET = True
GIT_TOKEN = None
GIT_USERNAME = None
GROWL_HOST = ""
GROWL_NOTIFY_ONDOWNLOAD = False
GROWL_NOTIFY_ONSNATCH = False
GROWL_NOTIFY_ONSUBTITLEDOWNLOAD = False
GROWL_PASSWORD = None
GUI_LANG = None
GUI_NAME = None
HANDLE_REVERSE_PROXY = False
HISTORY_LAYOUT = None
HISTORY_LIMIT = 0
HOME_LAYOUT = None
HTTPS_CERT = None
HTTPS_KEY = None
IGNORE_BROKEN_SYMLINKS = False
IGNORE_WORDS = "german,french,core2hd,dutch,swedish,reenc,MrLss"
IGNORED_SUBS_LIST = "dk,fin,heb,kor,nor,nordic,pl,swe"
IMAGE_CACHE = None
INDEXER_DEFAULT = 1
INDEXER_DEFAULT_LANGUAGE = None
INDEXER_TIMEOUT = None
INIT_LOCK = Lock()
ITASA_PASS = None
ITASA_USER = None
JOIN_APIKEY = ""
JOIN_ID = ""
JOIN_NOTIFY_ONDOWNLOAD = False
JOIN_NOTIFY_ONSNATCH = False
JOIN_NOTIFY_ONSUBTITLEDOWNLOAD = False
KEEP_PROCESSED_DIR = False
KODI_ALWAYS_ON = True
KODI_HOST = ""
KODI_NOTIFY_ONDOWNLOAD = False
KODI_NOTIFY_ONSNATCH = False
KODI_NOTIFY_ONSUBTITLEDOWNLOAD = False
KODI_PASSWORD = None
KODI_UPDATE_FULL = False
KODI_UPDATE_LIBRARY = False
KODI_UPDATE_ONLYFIRST = False
KODI_USERNAME = None
LAUNCH_BROWSER = False
LEGENDASTV_PASS = None
LEGENDASTV_USER = None
LIBNOTIFY_NOTIFY_ONDOWNLOAD = False
LIBNOTIFY_NOTIFY_ONPOSTPROCESS = False
LIBNOTIFY_NOTIFY_ONSNATCH = False
LIBNOTIFY_NOTIFY_ONSUBTITLEDOWNLOAD = False
LOCALHOST_IP = None
LOG_DIR = None
LOG_NR = 5
LOG_SIZE = 10.0
LOGO_URL = "https://sickchill.github.io/images/ico/favicon-64.png"
MATRIX_API_TOKEN = None
MATRIX_NOTIFY_DOWNLOAD = None
MATRIX_NOTIFY_SNATCH = None
MATRIX_NOTIFY_SUBTITLEDOWNLOAD = None
MATRIX_ROOM = None
MATRIX_SERVER = None
MESSAGES_LOCK = Lock()
METADATA_KODI = None
METADATA_MEDE8ER = None
METADATA_MEDIABROWSER = None
metadata_provider_dict = {}
METADATA_PS3 = None
METADATA_TIVO = None
METADATA_WDTV = None
MIN_AUTOPOSTPROCESSOR_FREQUENCY = 1
MIN_BACKLOG_FREQUENCY = 10
MIN_DAILYSEARCH_FREQUENCY = 10
MIN_UPDATE_FREQUENCY = 1
MOVE_ASSOCIATED_FILES = False
MY_ARGS = []
MY_FULLNAME = None
MY_NAME = None
NAMING_ABD_PATTERN = None
NAMING_ANIME = None
NAMING_ANIME_MULTI_EP = False
NAMING_ANIME_PATTERN = None
NAMING_CUSTOM_ABD = False
NAMING_CUSTOM_ANIME = False
NAMING_CUSTOM_SPORTS = False
NAMING_FORCE_FOLDERS = False
NAMING_MULTI_EP = False
NAMING_PATTERN = None
NAMING_SPORTS_PATTERN = None
NAMING_STRIP_YEAR = False
NEWS_LAST_READ = None
NEWS_LATEST = None
NEWS_UNREAD = 0
NEWS_URL = "https://sickchill.github.io/sickchill-news/news.md"
NEWZBIN = False
NEWZBIN_PASSWORD = None
NEWZBIN_USERNAME = None
NEWZNAB_DATA = None
newznabProviderList = []
NFO_RENAME = True
NMA_API = None
NMA_NOTIFY_ONDOWNLOAD = False
NMA_NOTIFY_ONSNATCH = False
NMA_NOTIFY_ONSUBTITLEDOWNLOAD = False
NMA_PRIORITY = 0
NMJ_DATABASE = None
NMJ_HOST = None
NMJ_MOUNT = None
NMJv2_DATABASE = None
NMJv2_DBLOC = None
NMJv2_HOST = None
NO_DELETE = False
NO_RESIZE = False
NO_RESTART = False
notificationsTaskScheduler = None
NOTIFY_ON_LOGIN = False
NOTIFY_ON_UPDATE = False
NZB_DIR = None
NZB_METHOD = None
NZBGET_CATEGORY = None
NZBGET_CATEGORY_ANIME = None
NZBGET_CATEGORY_ANIME_BACKLOG = None
NZBGET_CATEGORY_BACKLOG = None
NZBGET_HOST = None
NZBGET_PASSWORD = None
NZBGET_PRIORITY = 100
NZBGET_USE_HTTPS = False
NZBGET_USERNAME = None
NZBS = False
NZBS_HASH = None
NZBS_UID = None
OMGWTFNZBS = False
OMGWTFNZBS_APIKEY = None
OMGWTFNZBS_USERNAME = None
OPENSUBTITLES_PASS = None
OPENSUBTITLES_USER = None
PID = None
PIDFILE = ""
PLEX_CLIENT_HOST = None
PLEX_CLIENT_PASSWORD = None
PLEX_CLIENT_USERNAME = None
PLEX_NOTIFY_ONDOWNLOAD = False
PLEX_NOTIFY_ONSNATCH = False
PLEX_NOTIFY_ONSUBTITLEDOWNLOAD = False
PLEX_SERVER_HOST = None
PLEX_SERVER_HTTPS = None
PLEX_SERVER_PASSWORD = None
PLEX_SERVER_TOKEN = None
PLEX_SERVER_USERNAME = None
PLEX_UPDATE_LIBRARY = False
POSTER_SORTBY = None
POSTER_SORTDIR = None
POSTPONE_IF_SYNC_FILES = True
postProcessorTaskScheduler = None
PREFER_WORDS = ""
PROCESS_AUTOMATICALLY = False
PROCESS_METHOD = None
PROCESSOR_FOLLOW_SYMLINKS = False
PROG_DIR = sickchill_dir()
properFinderScheduler = None
PROVIDER_ORDER = []
providerList = []
PROWL_API = None
PROWL_MESSAGE_TITLE = "SickChill"
PROWL_NOTIFY_ONDOWNLOAD = False
PROWL_NOTIFY_ONSNATCH = False
PROWL_NOTIFY_ONSUBTITLEDOWNLOAD = False
PROWL_PRIORITY = 0
PROXY_INDEXERS = False
PROXY_SETTING = None
PUSHALOT_AUTHORIZATIONTOKEN = None
PUSHALOT_NOTIFY_ONDOWNLOAD = False
PUSHALOT_NOTIFY_ONSNATCH = False
PUSHALOT_NOTIFY_ONSUBTITLEDOWNLOAD = False
PUSHBULLET_API = None
PUSHBULLET_CHANNEL = None
PUSHBULLET_DEVICE = None
PUSHBULLET_NOTIFY_ONDOWNLOAD = False
PUSHBULLET_NOTIFY_ONSNATCH = False
PUSHBULLET_NOTIFY_ONSUBTITLEDOWNLOAD = False
PUSHOVER_APIKEY = None
PUSHOVER_DEVICE = None
PUSHOVER_NOTIFY_ONDOWNLOAD = False
PUSHOVER_NOTIFY_ONSNATCH = False
PUSHOVER_NOTIFY_ONSUBTITLEDOWNLOAD = False
PUSHOVER_PRIORITY = 0
PUSHOVER_SOUND = None
PUSHOVER_USERKEY = None
PYTIVO_HOST = ""
PYTIVO_NOTIFY_ONDOWNLOAD = False
PYTIVO_NOTIFY_ONSNATCH = False
PYTIVO_NOTIFY_ONSUBTITLEDOWNLOAD = False
PYTIVO_SHARE_NAME = ""
PYTIVO_TIVO_NAME = ""
PYTIVO_UPDATE_LIBRARY = False
QUALITY_ALLOW_HEVC = None
QUALITY_DEFAULT = SD
RANDOMIZE_PROVIDERS = False
RENAME_EPISODES = False
REQUIRE_WORDS = ""
ROCKETCHAT_ICON_EMOJI = None
ROCKETCHAT_NOTIFY_DOWNLOAD = None
ROCKETCHAT_NOTIFY_SNATCH = None
ROCKETCHAT_NOTIFY_SUBTITLEDOWNLOAD = None
ROCKETCHAT_WEBHOOK = None
ROOT_DIRS = None
SAB_APIKEY = None
SAB_CATEGORY = None
SAB_CATEGORY_ANIME = None
SAB_CATEGORY_ANIME_BACKLOG = None
SAB_CATEGORY_BACKLOG = None
SAB_FORCED = False
SAB_HOST = ""
SAB_PASSWORD = None
SAB_USERNAME = None
SCENE_DEFAULT = False
searchQueueScheduler = None
SEASON_FOLDERS_DEFAULT = False
showList = []
showQueueScheduler = None
SHOWS_RECENT = []
SHOWUPDATE_HOUR = None
showUpdateScheduler = None
SICKCHILL_BACKGROUND = None
SICKCHILL_BACKGROUND_PATH = None
SITE_MESSAGES = {}
SKIP_REMOVED_FILES = False
SLACK_ICON_EMOJI = None
SLACK_NOTIFY_DOWNLOAD = None
SLACK_NOTIFY_SNATCH = None
SLACK_NOTIFY_SUBTITLEDOWNLOAD = None
SLACK_WEBHOOK = None
SOCKET_TIMEOUT = None
SORT_ARTICLE = False
SSL_VERIFY = True
started = {}
STATUS_DEFAULT = None
STATUS_DEFAULT_AFTER = None
SUBSCENTER_PASS = None
SUBSCENTER_USER = None
SUBTITLES_DEFAULT = False
SUBTITLES_DIR = ""
SUBTITLES_EXTRA_SCRIPTS = []
SUBTITLES_FINDER_FREQUENCY = 1
SUBTITLES_HEARING_IMPAIRED = False
SUBTITLES_HISTORY = False
SUBTITLES_INCLUDE_SPECIALS = True
SUBTITLES_KEEP_ONLY_WANTED = False
SUBTITLES_LANGUAGES = []
SUBTITLES_MULTI = False
SUBTITLES_PERFECT_MATCH = False
SUBTITLES_SERVICES_ENABLED = []
SUBTITLES_SERVICES_LIST = []
subtitlesFinderScheduler = None
SYNC_FILES = "!sync,lftp-pget-status,bts,!qb,!qB"
SYNOLOGY_DSM_HOST = None
SYNOLOGY_DSM_PASSWORD = None
SYNOLOGY_DSM_PATH = None
SYNOLOGY_DSM_USERNAME = None
SYNOLOGYNOTIFIER_NOTIFY_ONDOWNLOAD = False
SYNOLOGYNOTIFIER_NOTIFY_ONSNATCH = False
SYNOLOGYNOTIFIER_NOTIFY_ONSUBTITLEDOWNLOAD = False
TELEGRAM_APIKEY = ""
TELEGRAM_ID = ""
TELEGRAM_NOTIFY_ONDOWNLOAD = False
TELEGRAM_NOTIFY_ONSNATCH = False
TELEGRAM_NOTIFY_ONSUBTITLEDOWNLOAD = False
THEME_NAME = None
TIME_PRESET = None
TIME_PRESET_W_SECONDS = None
TIMEZONE_DISPLAY = None
TMDB_API_KEY = "edc5f123313769de83a71e157758030b"
TORRENT_AUTH_TYPE = "none"
TORRENT_DIR = None
TORRENT_HIGH_BANDWIDTH = False
TORRENT_HOST = ""
TORRENT_LABEL = ""
TORRENT_LABEL_ANIME = ""
TORRENT_METHOD = None
TORRENT_PASSWORD = None
TORRENT_PATH = ""
TORRENT_PATH_INCOMPLETE = ""
TORRENT_PAUSED = False
TORRENT_RPCURL = "transmission"
TORRENT_SEED_TIME = None
TORRENT_USERNAME = None
TORRENT_VERIFY_CERT = False
torrentRssProviderList = []
TRACKERS_LIST = "udp://coppersurfer.tk:6969/announce,udp://open.demonii.com:1337,"
TRACKERS_LIST += "udp://9.rarbg.to:2710/announce"
TRACKERS_LIST += "udp://exodus.desync.com:6969,udp://9.rarbg.me:2710/announce,"
TRACKERS_LIST += "udp://glotorrents.pw:6969/announce,udp://tracker.openbittorrent.com:80/announce,"
TRACKERS_LIST += "udp://tracker.opentrackr.org:1337/announce,udp://tracker.internetwarriors.net:1337"
TRAKT_ACCESS_TOKEN = None
# TRAKT_API_KEY = 'd4161a7a106424551add171e5470112e4afdaf2438e6ef2fe0548edc75924868'
TRAKT_API_KEY = "5c65f55e11d48c35385d9e8670615763a605fad28374c8ae553a7b7a50651ddd"
TRAKT_API_SECRET = "b53e32045ac122a445ef163e6d859403301ffe9b17fb8321d428531b69022a82"
TRAKT_API_URL = "https://api-v2launch.trakt.tv/"
TRAKT_BLACKLIST_NAME = None
TRAKT_DEFAULT_INDEXER = None
TRAKT_METHOD_ADD = None
TRAKT_OAUTH_URL = "https://trakt.tv/"
TRAKT_PIN_URL = "https://trakt.tv/pin/4562"
TRAKT_REFRESH_TOKEN = None
TRAKT_REMOVE_SERIESLIST = False
TRAKT_REMOVE_SHOW_FROM_SICKCHILL = False
TRAKT_REMOVE_WATCHLIST = False
TRAKT_START_PAUSED = False
TRAKT_SYNC = False
TRAKT_SYNC_REMOVE = False
TRAKT_SYNC_WATCHLIST = False
TRAKT_TIMEOUT = None
TRAKT_USE_RECOMMENDED = False
TRAKT_USERNAME = None
traktCheckerScheduler = None
TRASH_REMOVE_SHOW = False
TRASH_ROTATE_LOGS = False
TRIM_ZERO = False
TV_DOWNLOAD_DIR = None
TVDB_USER = None
TVDB_USER_KEY = None
TWILIO_ACCOUNT_SID = ""
TWILIO_AUTH_TOKEN = ""
TWILIO_NOTIFY_ONDOWNLOAD = False
TWILIO_NOTIFY_ONSNATCH = False
TWILIO_NOTIFY_ONSUBTITLEDOWNLOAD = False
TWILIO_PHONE_SID = ""
TWILIO_TO_NUMBER = ""
TWITTER_DMTO = None
TWITTER_NOTIFY_ONDOWNLOAD = False
TWITTER_NOTIFY_ONSNATCH = False
TWITTER_NOTIFY_ONSUBTITLEDOWNLOAD = False
TWITTER_PASSWORD = None
TWITTER_PREFIX = None
TWITTER_USEDM = False
TWITTER_USERNAME = None
UNAR_TOOL = rarfile.UNAR_TOOL
UNPACK_DISABLED = 0
UNPACK = UNPACK_DISABLED
UNPACK_DIR = ""
UNPACK_PROCESS_CONTENTS = 1
UNPACK_PROCESS_INTACT = 2
UNRAR_TOOL = rarfile.UNRAR_TOOL
UPDATE_FREQUENCY = None
USE_ANIDB = False
USE_BOXCAR2 = False
USE_DISCORD = False
USE_EMAIL = False
USE_EMBY = False
USE_FAILED_DOWNLOADS = False
USE_FREE_SPACE_CHECK = True
USE_FREEMOBILE = False
USE_GROWL = False
USE_ICACLS = True
USE_JOIN = False
USE_KODI = False
USE_LIBNOTIFY = False
USE_LISTVIEW = False
USE_MATRIX = False
USE_NMA = False
USE_NMJ = False
USE_NMJv2 = False
USE_NZBS = False
USE_PLEX_CLIENT = False
USE_PLEX_SERVER = False
USE_PROWL = False
USE_PUSHALOT = False
USE_PUSHBULLET = False
USE_PUSHOVER = False
USE_PYTIVO = False
USE_ROCKETCHAT = False
USE_SLACK = False
USE_SUBTITLES = False
USE_SYNOINDEX = False
USE_SYNOLOGYNOTIFIER = False
USE_TELEGRAM = False
USE_TORRENTS = False
USE_TRAKT = False
USE_TWILIO = False
USE_TWITTER = False
USENET_RETENTION = None
VERSION_NOTIFY = False
versionCheckScheduler = None
WEB_COOKIE_SECRET = None
WEB_HOST = None
WEB_IPV6 = False
WEB_LOG = False
WEB_PASSWORD = None
WEB_PORT = None
WEB_ROOT = None
WEB_USE_GZIP = True
WEB_USERNAME = None
WINDOWS_SHARES = {}

movie_list: "MovieList" = None


def get_backlog_cycle_time():
    cycletime = DAILYSEARCH_FREQUENCY * 2 + 7
    return max([cycletime, 720])


unpackStrings = NumDict(
    {
        UNPACK_DISABLED: _("Ignore (do not process contents)"),
        UNPACK_PROCESS_CONTENTS: _("Unpack (process contents)"),
        UNPACK_PROCESS_INTACT: _("Treat as video (process archive as-is)"),
    }
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2016-5-26

@author: HAKASHUN
'''

LOCALE = [
    [u"たった今", u"すぐに"],
    [u"%s秒前", u"%s秒以内"],
    [u"1分前", u"1分以内"],
    [u"%s分前", u"%s分以内"],
    [u"1時間前", u"1時間以内"],
    [u"%s時間前", u"%s時間以内"],
    [u"昨日", u"1日以内"],
    [u"%s日前", u"%s日以内"],
    [u"先週", u"1週間以内"],
    [u"%s週間前", u"%s週間以内"],
    [u"1ヶ月前", u"1ヶ月以内"],
    [u"%sヶ月前", u"%sヶ月以内"],
    [u"1年前", u"1年以内"],
    [u"%s年前", u"%s年以内"],
]


import datetime
import re

from dateutil import tz

from sickchill.helper.common import try_int

from .. import logger
from . import db, helpers

# regex to parse time (12/24 hour format)
time_regex = re.compile(r"(?P<hour>\d{1,2})(?:[:.](?P<minute>\d{2})?)? ?(?P<meridiem>[PA]\.? ?M?)?\b", re.I)

network_dict = {}

sb_timezone = tz.gettz()

missing_network_timezones = set()


def update_network_dict():
    """Update timezone information from SR repositories"""

    url = "https://sickchill.github.io/sb_network_timezones/network_timezones.txt"
    data = helpers.getURL(url, session=helpers.make_session(), returns="text")
    if not data:
        logger.warning("Updating network timezones failed, this can happen from time to time. URL: {0}".format(url))
        load_network_dict()
        return

    d = {}
    try:
        for line in data.splitlines():
            (key, val) = line.strip().rsplit(":", 1)
            if not val:  # n Lets set a default
                val = "US/Eastern"
            if key and val:
                d[key.lower()] = val
    except (IOError, OSError):
        pass

    if not d:
        logger.warning("Parsing network timezones failed, not going to touch the db")
        load_network_dict()
        return

    cache_db_con = db.DBConnection("cache.db")

    network_list = dict(cache_db_con.select("SELECT * FROM network_timezones;"))

    queries = []
    for network, timezone in d.items():
        existing = network in network_list
        if not existing:
            queries.append(["INSERT OR IGNORE INTO network_timezones VALUES (?,?);", [network, timezone]])
        elif network_list[network] != timezone:
            queries.append(["UPDATE OR IGNORE network_timezones SET timezone = ? WHERE network_name = ?;", [timezone, network]])

        if existing:
            del network_list[network]

    for network in network_list:
        queries.append(["DELETE FROM network_timezones WHERE network_name = ?;", [network]])

    if queries:
        cache_db_con.mass_action(queries)
        load_network_dict()


def load_network_dict():
    """
    Load network timezones from db into dict network_dict (global dict)
    """
    try:
        cache_db_con = db.DBConnection("cache.db")
        cur_network_list = cache_db_con.select("SELECT * FROM network_timezones;")
        if not cur_network_list:
            update_network_dict()
            cur_network_list = cache_db_con.select("SELECT * FROM network_timezones;")

        network_dict.clear()
        network_dict.update(dict(cur_network_list))
    except Exception:
        pass


def get_network_timezone(network):
    """
    Get the timezone of a network, or return sb_timezone

    :param network: network to look up
    :return: network timezone if found, or sb_timezone
    """

    orig_network = network
    if network:
        network = network.lower()

    network_tz_name = network_dict.get(network)
    if network and not (network_tz_name or network in missing_network_timezones):
        missing_network_timezones.add(network)
        severity = (logger.ERROR, logger.INFO)[network.endswith("none")]
        logger.log(
            severity,
            _(
                "Missing time zone for network: {orig_network}. Check valid network is set in indexer (theTVDB) before filing issue.".format(
                    orig_network=orig_network
                )
            ),
        )

    try:
        network_tz = (tz.gettz(network_tz_name) or sb_timezone) if network_tz_name else sb_timezone
    except Exception:
        return sb_timezone
    return network_tz


def parse_date_time(d, t, network):
    """
    Parse date and time string into local time

    :param d: date string
    :param t: time string
    :param network: network to use as base
    :return: datetime object containing local time
    """

    if not network_dict:
        load_network_dict()

    parsed_time = time_regex.search(t)
    network_tz = get_network_timezone(network)

    hr = 0
    m = 0

    if parsed_time:
        hr = try_int(parsed_time.group("hour"))
        m = try_int(parsed_time.group("minute"))

        ap = parsed_time.group("meridiem")
        ap = ap[0].lower() if ap else ""

        if ap == "a" and hr == 12:
            hr -= 12
        elif ap == "p" and hr != 12:
            hr += 12

        hr = hr if 0 <= hr <= 23 else 0
        m = m if 0 <= m <= 59 else 0

    result = datetime.datetime.fromordinal(max(try_int(d), 1))

    return result.replace(hour=hr, minute=m, tzinfo=network_tz)


def test_timeformat(time_string):
    return time_regex.search(time_string) is not None



# -*- coding: utf-8 -*-

############################ Copyrights and license ############################
#                                                                              #
# Copyright 2020 Raju Subramanian <coder@mahesh.net>                           #
#                                                                              #
# This file is part of PyGithub.                                               #
# http://pygithub.readthedocs.io/                                              #
#                                                                              #
# PyGithub is free software: you can redistribute it and/or modify it under    #
# the terms of the GNU Lesser General Public License as published by the Free  #
# Software Foundation, either version 3 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# PyGithub is distributed in the hope that it will be useful, but WITHOUT ANY  #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS    #
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more #
# details.                                                                     #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with PyGithub. If not, see <http://www.gnu.org/licenses/>.             #
#                                                                              #
################################################################################

import github.GithubObject
import github.NamedUser


class GithubApp(github.GithubObject.CompletableGithubObject):
    """
    This class represents github apps. The reference can be found here https://docs.github.com/en/rest/reference/apps
    """

    def __repr__(self):
        return self.get__repr__({"id": self._id.value, "url": self._url.value})

    @property
    def created_at(self):
        """
        :type: datetime.datetime
        """
        self._completeIfNotSet(self._created_at)
        return self._created_at.value

    @property
    def description(self):
        """
        :type: string
        """
        self._completeIfNotSet(self._description)
        return self._description.value

    @property
    def events(self):
        """
        :type: list of string
        """
        self._completeIfNotSet(self._events)
        return self._events.value

    @property
    def external_url(self):
        """
        :type: string
        """
        self._completeIfNotSet(self._external_url)
        return self._external_url.value

    @property
    def html_url(self):
        """
        :type: string
        """
        self._completeIfNotSet(self._html_url)
        return self._html_url.value

    @property
    def id(self):
        """
        :type: int
        """
        self._completeIfNotSet(self._id)
        return self._id.value

    @property
    def name(self):
        """
        :type: string
        """
        self._completeIfNotSet(self._name)
        return self._name.value

    @property
    def owner(self):
        """
        :type: :class:`Github.NamedUser.NamedUser`
        """
        self._completeIfNotSet(self._owner)
        return self._owner.value

    @property
    def permissions(self):
        """
        :type: dict
        """
        self._completeIfNotSet(self._permissions)
        return self._permissions.value

    @property
    def slug(self):
        """
        :type: string
        """
        return self._slug.value

    @property
    def updated_at(self):
        """
        :type: datetime.datetime
        """
        self._completeIfNotSet(self._updated_at)
        return self._updated_at.value

    @property
    def url(self):
        """
        :type: string
        """
        return self._url.value

    def _initAttributes(self):
        self._created_at = github.GithubObject.NotSet
        self._description = github.GithubObject.NotSet
        self._events = github.GithubObject.NotSet
        self._external_url = github.GithubObject.NotSet
        self._html_url = github.GithubObject.NotSet
        self._id = github.GithubObject.NotSet
        self._name = github.GithubObject.NotSet
        self._owner = github.GithubObject.NotSet
        self._permissions = github.GithubObject.NotSet
        self._slug = github.GithubObject.NotSet
        self._updated_at = github.GithubObject.NotSet
        self._url = github.GithubObject.NotSet

    def _useAttributes(self, attributes):
        if "created_at" in attributes:  # pragma no branch
            self._created_at = self._makeDatetimeAttribute(attributes["created_at"])
        if "description" in attributes:  # pragma no branch
            self._description = self._makeStringAttribute(attributes["description"])
        if "events" in attributes:  # pragma no branch
            self._events = self._makeListOfStringsAttribute(attributes["events"])
        if "external_url" in attributes:  # pragma no branch
            self._external_url = self._makeStringAttribute(attributes["external_url"])
        if "html_url" in attributes:  # pragma no branch
            self._html_url = self._makeStringAttribute(attributes["html_url"])
        if "id" in attributes:  # pragma no branch
            self._id = self._makeIntAttribute(attributes["id"])
        if "name" in attributes:  # pragma no branch
            self._name = self._makeStringAttribute(attributes["name"])
        if "owner" in attributes:  # pragma no branch
            self._owner = self._makeClassAttribute(
                github.NamedUser.NamedUser, attributes["owner"]
            )
        if "permissions" in attributes:  # pragma no branch
            self._permissions = self._makeDictAttribute(attributes["permissions"])
        if "slug" in attributes:  # pragma no branch
            self._slug = self._makeStringAttribute(attributes["slug"])
            self._url = self._makeStringAttribute("/apps/" + attributes["slug"])
        if "updated_at" in attributes:  # pragma no branch
            self._updated_at = self._makeDatetimeAttribute(attributes["updated_at"])
        if "url" in attributes:
            self._url = self._makeStringAttribute(attributes["url"])



# -*- coding: utf-8 -*-
class Error(Exception):
    """Base class for exceptions in subliminal."""
    pass


class ProviderError(Error):
    """Exception raised by providers."""
    pass


class ConfigurationError(ProviderError):
    """Exception raised by providers when badly configured."""
    pass


class AuthenticationError(ProviderError):
    """Exception raised by providers when authentication failed."""
    pass


class ServiceUnavailable(ProviderError):
    """Exception raised when status is '503 Service Unavailable'."""
    pass


class DownloadLimitExceeded(ProviderError):
    """Exception raised by providers when download limit is exceeded."""
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2019-12-19

@author: ANJD
'''

base = [
    ["الآن", "خلال لحظات"],
    ["منذ ثانيتين", "خلال ثانيتين", "منذ %s ثواني", "خلال %s ثواني", "منذ %s ثانية", "خلال %s ثانية"],
    ["منذ دقيقة", "خلال دقيقة"],
    ["منذ دقيقتين", "خلال دقيقتين", "منذ %s دقائق", "خلال %s دقائق", "منذ %s دقيقة", "خلال %s دقيقة"],
    ["منذ ساعة", "خلال ساعة"],
    ["منذ ساعتين", "خلال ساعتين", "منذ %s ساعات", "خلال %s ساعات", "منذ %s ساعة", "خلال %s ساعة"],
    ["منذ يوم", "خلال يوم"],
    ["منذ يومين", "خلال يومين", "منذ %s أيام", "خلال %s أيام", "منذ %s يوم", "خلال %s يوم"],
    ["منذ أسبوع", "خلال أسبوع"],
    ["منذ أسبوعين", "خلال أسبوعين", "منذ %s أسابيع", "خلال %s أسابيع", "منذ %s أسبوع", "خلال %s أسبوع"],
    ["منذ شهر", "خلال شهر"],
    ["منذ شهرين", "خلال شهرين", "منذ %s أشهر", "خلال %s أشهر", "منذ %s شهراً", "خلال %s شهراً"],
    ["منذ سنة", "خلال سنة"],
    ["منذ سنتين", "خلال سنتين", "منذ %s سنوات", "خلال %s سنوات", "منذ %s سنة", "خلال %s سنة"]
]


def generate(row, y):
    def formatting(time):
        if row % 2 == 0:
            return base[row][y]

        if time == 2:
            return base[row][y]

        if time in range(3, 11):
            return base[row][y + 2]

        return base[row][y + 4]

    return formatting


LOCALE = generate



# -*- coding: utf-8 -*-

"""
This module implements the Language functionality of TheTVDb API.
Allows to retrieve the languages list and info.

See [Languages API section](https://api.thetvdb.com/swagger#!/Languages)
"""
import sys
if sys.version_info > (2, 8):
    from builtins import dict
from .base import TVDB

class Languages(TVDB):
    """
    Languages class to retrieve the languages list and info.
    """
    _BASE_PATH = 'languages'
    _URLS = {
        'all': '',
        'language': '/{lid}'
    }
    LANGUAGES = {}
    _ALL_PARSED = False

    def all(self):
        """
        Get the full languages list and set it to all attribute.

        Returns a list of languages with their info.

        For example

            #!python
            >>> import tvdbsimple as tvdb
            >>> tvdb.KEYS.API_KEY = 'YOUR_API_KEY'
            >>> lng = tvdb.Languages()
            >>> response = lng.all()
            >>> lng.all[0]['englishName']
            'Chinese'

        """
        if self._ALL_PARSED:
            return self.LANGUAGES.values()
            
        path = self._get_path('all')
        
        response = self._GET(path)
        self._set_attrs_to_values({'all': response})
        for lang in response:
            if 'id' in lang:
                self.LANGUAGES[lang['id']] = lang
        self._ALL_PARSED = True
        return response

    def language(self, id):
        """
        Get the language info for a specific language id.
        
        `id` is the specific language id to retrieve.

        Returns a dict rwith all the language info.

        For example

            #!python
            >>> import tvdbsimple as tvdb
            >>> tvdb.KEYS.API_KEY = 'YOUR_API_KEY'
            >>> lng = tvdb.Languages()
            >>> response = lng.language(7)
            >>> response['englishName']
            'English'

        """
        if id in self.LANGUAGES:
            return self.LANGUAGES[id]

        path = self._get_path('language').format(lid=id)
        
        response = self._GET(path)
        if 'id' in response:
            self.LANGUAGES[response['id']] = response
        return response

    def __iter__(self):
        for i in self.all():
            yield i

    def __getitem__(self, id):
        return self.language(id)

    def __setitem__(self):
        raise Exception('Function Disabled')
        
    def __delitem__(self):
        raise Exception('Function Disabled')



"""
Test the SR API
"""

import unittest


class APITestEpisodes(unittest.TestCase):
    """
    Test episode commands for API

    This tests all episode related commands from the legacy api.
    """

    @unittest.skip("Not yet implemented")
    def test_episode(self):
        """
        Test getting detailed information about an episode using the legacy API.

        :return: None
        """
        pass

    @unittest.skip("Not yet implemented")
    def test_episode_search(self):
        """
        Test searching for an episode using the legacy API.
        The response might take some time.
        :return: None
        """
        pass

    @unittest.skip("Not yet implemented")
    def test_episode_set_status(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_episode_subtitle_search(self):
        pass


class APITestShows(unittest.TestCase):
    """
    Test shows commands for API
    """

    @unittest.skip("Not yet implemented")
    def test_shows(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_shows_stats(self):
        pass


class APITestShow(unittest.TestCase):
    """
    Test show commands for API
    """

    @unittest.skip("Not yet implemented")
    def test_show(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_show_add_existing(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_show_add_new(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_show_cache(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_show_delete(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_show_get_banner(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_show_get_fanart(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_show_get_network_logo(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_show_get_poster(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_show_get_quality(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_show_pause(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_show_refresh(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_show_season_list(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_show_seasons(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_show_set_quality(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_show_stats(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_show_update(self):
        pass


class APITestSickChill(unittest.TestCase):
    @unittest.skip("Not yet implemented")
    def test_sickchill(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_sb_add_root_dir(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_sb_check_scheduler(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_sb_check_version(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_sb_delete_root_dir(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_sb_get_defaults(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_sb_get_messages(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_sb_get_root_dirs(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_sb_pause_backlog(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_sb_ping(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_sb_restart(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_sb_search_indexers(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_sb_search_tvdb(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_sb_search_tvrage(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_sb_set_defaults(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_sb_shutdown(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_sb_update(self):
        pass


class APITestHistory(unittest.TestCase):
    @unittest.skip("Not yet implemented")
    def test_history(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_history_clear(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_history_trim(self):
        pass


class APITestMisc(unittest.TestCase):
    @unittest.skip("Not yet implemented")
    def test_backlog(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_exceptions(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_failed(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_future(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_help(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_logs(self):
        pass

    @unittest.skip("Not yet implemented")
    def test_post_process(self):
        pass


TEST_CLASSES = {APITestEpisodes, APITestHistory, APITestMisc, APITestShow, APITestShows, APITestSickChill}


def run_all():
    """
    Run all tests
    :return:
    """
    unittest.main(verbosity=2)


if __name__ == "__main__":
    run_all()



import os.path
from mimetypes import guess_type

import imagesize

import sickchill
from sickchill import logger, settings
from sickchill.helper.exceptions import ShowDirectoryNotFoundException
from sickchill.providers.metadata.generic import GenericMetadata
from sickchill.providers.metadata.helpers import getShowImage

from . import helpers


class ImageCache(object):
    def __init__(self):
        pass

    def __del__(self):
        pass

    @staticmethod
    def _cache_dir():
        """
        Builds up the full path to the image cache directory
        """
        return os.path.abspath(os.path.join(settings.CACHE_DIR, "images"))

    def _thumbnails_dir(self):
        """
        Builds up the full path to the thumbnails image cache directory
        """
        return os.path.abspath(os.path.join(self._cache_dir(), "thumbnails"))

    def poster_path(self, indexer_id):
        """
        Builds up the path to a poster cache for a given Indexer ID

        :param indexer_id: ID of the show to use in the file name
        :return: a full path to the cached poster file for the given Indexer ID
        """
        poster_filename = str(indexer_id) + ".poster.jpg"
        return os.path.join(self._cache_dir(), poster_filename)

    def banner_path(self, indexer_id):
        """
        Builds up the path to a banner cache for a given Indexer ID

        :param indexer_id: ID of the show to use in the file name
        :return: a full path to the cached banner file for the given Indexer ID
        """
        banner_filename = str(indexer_id) + ".banner.jpg"
        return os.path.join(self._cache_dir(), banner_filename)

    def fanart_path(self, indexer_id):
        """
        Builds up the path to a fanart cache for a given Indexer ID

        :param indexer_id: ID of the show to use in the file name
        :return: a full path to the cached fanart file for the given Indexer ID
        """
        fanart_filename = str(indexer_id) + ".fanart.jpg"
        return os.path.join(self._cache_dir(), fanart_filename)

    def poster_thumb_path(self, indexer_id):
        """
        Builds up the path to a poster thumb cache for a given Indexer ID

        :param indexer_id: ID of the show to use in the file name
        :return: a full path to the cached poster thumb file for the given Indexer ID
        """
        posterthumb_filename = str(indexer_id) + ".poster.jpg"
        return os.path.join(self._thumbnails_dir(), posterthumb_filename)

    def banner_thumb_path(self, indexer_id):
        """
        Builds up the path to a banner thumb cache for a given Indexer ID

        :param indexer_id: ID of the show to use in the file name
        :return: a full path to the cached banner thumb file for the given Indexer ID
        """
        bannerthumb_filename = str(indexer_id) + ".banner.jpg"
        return os.path.join(self._thumbnails_dir(), bannerthumb_filename)

    def has_poster(self, indexer_id):
        """
        Returns true if a cached poster exists for the given Indexer ID
        """
        poster_path = self.poster_path(indexer_id)
        logger.debug("Checking if file " + str(poster_path) + " exists")
        return os.path.isfile(poster_path)

    def has_banner(self, indexer_id):
        """
        Returns true if a cached banner exists for the given Indexer ID
        """
        banner_path = self.banner_path(indexer_id)
        logger.debug("Checking if file " + str(banner_path) + " exists")
        return os.path.isfile(banner_path)

    def has_fanart(self, indexer_id):
        """
        Returns true if a cached fanart exists for the given Indexer ID
        """
        fanart_path = self.fanart_path(indexer_id)
        logger.debug("Checking if file " + str(fanart_path) + " exists")
        return os.path.isfile(fanart_path)

    def has_poster_thumb(self, indexer_id):
        """
        Returns true if a cached poster thumbnail exists for the given Indexer ID
        """
        poster_thumb_path = self.poster_thumb_path(indexer_id)
        logger.debug("Checking if file " + str(poster_thumb_path) + " exists")
        return os.path.isfile(poster_thumb_path)

    def has_banner_thumb(self, indexer_id):
        """
        Returns true if a cached banner exists for the given Indexer ID
        """
        banner_thumb_path = self.banner_thumb_path(indexer_id)
        logger.debug("Checking if file " + str(banner_thumb_path) + " exists")
        return os.path.isfile(banner_thumb_path)

    def image_url(self, indexer_id, which):
        path = self.__getattribute__(which + "_path")(indexer_id)
        if os.path.isfile(path):
            return "cache" + path.split(settings.CACHE_DIR)[1].replace("\\", "/")
        if which == "fanart" and settings.SICKCHILL_BACKGROUND:
            return "ui/sickchill_background"

        return ("images/poster.png", "images/banner.png")["banner" in which]

    BANNER = 1
    POSTER = 2
    BANNER_THUMB = 3
    POSTER_THUMB = 4
    FANART = 5

    image_str = {BANNER: "banner", BANNER_THUMB: "banner thumbnail", POSTER: "poster", POSTER_THUMB: "poster thumbnail", FANART: "fanart"}

    def which_type(self, path):
        """
        Analyzes the image provided and attempts to determine whether it is a poster or banner.

        :param path: full path to the image
        :return: BANNER, POSTER if it concluded one or the other, or None if the image was neither (or didn't exist)
        """

        if not os.path.isfile(path):
            logger.warning("Couldn't check the type of " + str(path) + " cause it doesn't exist")
            return None

        width, height = imagesize.get(path)

        if not (width and height):
            logger.debug("Unable to get metadata from " + str(path) + ", not using your existing image")
            return None

        img_ratio = float(width) / float(height)

        # most posters are around 0.68 width/height ratio (eg. 680/1000)
        if 0.55 < img_ratio < 0.8:
            return self.POSTER

        # most banners are around 5.4 width/height ratio (eg. 758/140)
        elif 5 < img_ratio < 6:
            return self.BANNER

        # most fanart are around 1.77777 width/height ratio (eg. 1280/720 and 1920/1080)
        elif 1.7 < img_ratio < 1.8:
            return self.FANART
        else:
            logger.warning("Image has size ratio of " + str(img_ratio) + ", unknown type")
            return None

    @staticmethod
    def image_data(path):
        """
        :return: The content of the desired media file
        """

        if os.path.isfile(path):
            with open(path, "rb") as content:
                return content.read()

        return None

    @staticmethod
    def content_type(path):
        """
        :return: The mime type of the current media
        """

        if os.path.isfile(path):
            return guess_type(path)[0]

        return ""

    def _cache_image_from_file(self, image_path, img_type, indexer_id):
        """
        Takes the image provided and copies it to the cache folder

        :param image_path: path to the image we're caching
        :param img_type: BANNER or POSTER or FANART
        :param indexer_id: id of the show this image belongs to
        :return: bool representing success
        """

        # generate the path based on the type & indexer_id
        if img_type == self.POSTER:
            dest_path = self.poster_path(indexer_id)
        elif img_type == self.BANNER:
            dest_path = self.banner_path(indexer_id)
        elif img_type == self.FANART:
            dest_path = self.fanart_path(indexer_id)
        else:
            logger.exception("Invalid cache image type: " + str(img_type))
            return False

        # make sure the cache folder exists before we try copying to it
        if not os.path.isdir(self._cache_dir()):
            logger.info("Image cache dir didn't exist, creating it at " + str(self._cache_dir()))
            os.makedirs(self._cache_dir())

        if not os.path.isdir(self._thumbnails_dir()):
            logger.info("Thumbnails cache dir didn't exist, creating it at " + str(self._thumbnails_dir()))
            os.makedirs(self._thumbnails_dir())

        logger.info("Copying from " + image_path + " to " + dest_path)
        helpers.copyFile(image_path, dest_path)

        return True

    def _cache_image_from_indexer(self, show_obj, img_type):
        """
        Retrieves an image of the type specified from indexer and saves it to the cache folder

        :param show_obj: TVShow object that we want to cache an image for
        :param img_type: BANNER or POSTER or FANART
        :return: bool representing success
        """

        metadata_generator = GenericMetadata()

        # generate the path based on the type & indexer_id
        if img_type == self.POSTER:
            img_url = sickchill.indexer.series_poster_url(show_obj)
            if not img_url:
                img_url = metadata_generator._retrieve_show_image_urls_from_fanart(show_obj, "poster")
            if not img_url:
                img_url = metadata_generator._retrieve_show_image_urls_from_tmdb(show_obj, "poster")
            dest_path = self.poster_path(show_obj.indexerid)
        elif img_type == self.BANNER:
            img_url = sickchill.indexer.series_banner_url(show_obj)
            if not img_url:
                img_url = metadata_generator._retrieve_show_image_urls_from_fanart(show_obj, "banner")
            dest_path = self.banner_path(show_obj.indexerid)
        elif img_type == self.POSTER_THUMB:
            img_url = sickchill.indexer.series_poster_url(show_obj, thumb=True)
            if not img_url:
                img_url = metadata_generator._retrieve_show_image_urls_from_fanart(show_obj, "poster", thumb=True)
            if not img_url:
                img_url = metadata_generator._retrieve_show_image_urls_from_tmdb(show_obj, "poster")
            dest_path = self.poster_thumb_path(show_obj.indexerid)
        elif img_type == self.BANNER_THUMB:
            img_url = sickchill.indexer.series_banner_url(show_obj, thumb=True)
            if not img_url:
                img_url = metadata_generator._retrieve_show_image_urls_from_fanart(show_obj, "banner", thumb=True)
            dest_path = self.banner_thumb_path(show_obj.indexerid)
        elif img_type == self.FANART:
            img_url = sickchill.indexer.series_fanart_url(show_obj)
            if not img_url:
                img_url = metadata_generator._retrieve_show_image_urls_from_fanart(show_obj, "fanart")
            if not img_url:
                img_url = metadata_generator._retrieve_show_image_urls_from_tmdb(show_obj, "fanart")
            dest_path = self.fanart_path(show_obj.indexerid)
        else:
            logger.exception("Invalid cache image type: " + str(img_type))
            return False

        # retrieve the image from indexer using the generic metadata class
        # TODO: refactor
        img_data = getShowImage(img_url)
        result = metadata_generator._write_image(img_data, dest_path)

        return result

    def fill_cache(self, show_obj):
        """
        Caches all images for the given show. Copies them from the show dir if possible, or
        downloads them from indexer if they aren't in the show dir.

        :param show_obj: TVShow object to cache images for
        """

        logger.debug("Checking if we need any cache images for show " + str(show_obj.indexerid))

        # check if the images are already cached or not
        need_images = {
            self.POSTER: not self.has_poster(show_obj.indexerid),
            self.BANNER: not self.has_banner(show_obj.indexerid),
            self.POSTER_THUMB: not self.has_poster_thumb(show_obj.indexerid),
            self.BANNER_THUMB: not self.has_banner_thumb(show_obj.indexerid),
            self.FANART: not self.has_fanart(show_obj.indexerid),
        }

        if not any(need_images.values()):
            logger.debug("No new cache images needed, not retrieving new ones")
            return

        # check the show dir for poster or banner images and use them
        if need_images[self.POSTER] or need_images[self.BANNER] or need_images[self.FANART]:
            try:
                for cur_provider in settings.metadata_provider_dict.values():
                    logger.debug("[{}] Checking if we can use images from {} metadata".format(show_obj.indexerid, cur_provider.name))

                    for method in (cur_provider.get_poster_path, cur_provider.get_banner_path, cur_provider.get_fanart_path):
                        current_path = method(show_obj)
                        if os.path.isfile(current_path):
                            cur_filename = os.path.abspath(current_path)
                            cur_file_type = self.which_type(cur_filename)

                            if cur_file_type is None:
                                logger.warning("Unable to retrieve image type, not using the image from " + str(cur_filename))
                                continue

                            logger.debug(
                                "[{}] Checking if {} ({}) needs cached: {}".format(
                                    show_obj.indexerid, cur_filename, self.image_str[cur_file_type], need_images[cur_file_type]
                                )
                            )

                            if cur_file_type in need_images and need_images[cur_file_type]:
                                logger.debug(
                                    "[{}] Found a {} in the show dir that doesn't exist in the cache, caching it".format(
                                        show_obj.indexerid, self.image_str[cur_file_type]
                                    )
                                )
                                self._cache_image_from_file(cur_filename, cur_file_type, show_obj.indexerid)
                                need_images[cur_file_type] = False

            except ShowDirectoryNotFoundException:
                logger.debug("[{}] Unable to search for images in show dir because it doesn't exist".format(show_obj.indexerid))

        # download from indexer for missing ones
        for cur_image_type in need_images:
            logger.debug("[{}] Seeing if we still need a {}: {}".format(show_obj.indexerid, self.image_str[cur_image_type], need_images[cur_image_type]))
            if cur_image_type in need_images and need_images[cur_image_type]:
                self._cache_image_from_indexer(show_obj, cur_image_type)

        logger.info("Done cache check")



class ShowResult(object):
    def __init__(self, title, first_aired, next_airs, last_aired, status, language, indexer, description):
        self.title = title
        self.first_aired = first_aired
        self.next_airs = next_airs
        self.last_aired = last_aired
        self.status = status
        self.language = language
        self.indexer = indexer
        self.description = description



import datetime


def valmap(from_list, to_list, index=0):
    assert len(from_list) == len(to_list)
    def inner(*args):
        args = list(args)
        to_list_i = from_list.index(args[index])
        args[index] = to_list[to_list_i]
        return args

    return inner

def choose(arg):
    def inner(*args):
        if isinstance(arg, int):
            return args[arg]
        elif isinstance(arg, (list, range)):
            return list([args[i] for i in arg])
        else:
            raise AttributeError('Unhandled type {0}'.format(type(arg)))

    return inner

def int_to_bool(arg):
    return arg in [1, '1']

def bool_to_int(arg):
    return 1 if arg else 0

def check_success(arg):
    return arg == 0

def us_to_datetime(arg):
    return to_datetime(arg)

def s_to_datetime(arg):
    # Return None if RTorrent returns a timestamp of zero
    if arg <= 0:
        return None
    
    # RTorrent timestamps are converted to UTC
    # utcfromtimestamp will return a generic datetime object
    # without a timezone associated with it
    return datetime.datetime.utcfromtimestamp(arg) \
        .replace(tzinfo=datetime.timezone.utc)

def to_datetime(arg):
    # Return None if RTorrent returns a timestamp of zero
    if arg <= 0:
        return None

    # RTorrent timestamps are in microseconds, we need them it in seconds
    arg /= 1.0E6

    # RTorrent timestamps are converted to UTC
    # utcfromtimestamp will return a generic datetime object
    # without a timezone associated with it
    return datetime.datetime.utcfromtimestamp(arg) \
        .replace(tzinfo=datetime.timezone.utc)



"""
Test exception logging
"""

import unittest

import pytest

from sickchill import logger


class IssueSubmitterBasicTests(unittest.TestCase):
    """
    Tests logging of exceptions
    """

    def test_submitter(self):
        """
        Test that an exception is raised
        """
        try:
            with pytest.raises(Exception):
                raise Exception("FAKE EXCEPTION")
        except Exception as error:
            logger.exception("FAKE ERROR: " + str(error))
            logger.submit_errors()
            with pytest.raises(Exception):
                raise


if __name__ == "__main__":
    print("==================")
    print("STARTING - ISSUE SUBMITTER TESTS")
    print("==================")
    print("######################################################################")

    SUITE = unittest.TestLoader().loadTestsFromTestCase(IssueSubmitterBasicTests)
    unittest.TextTestRunner(verbosity=2).run(SUITE)



"""
Test shows
"""


import unittest

from sickchill import settings
from sickchill.helper.exceptions import MultipleShowObjectsException
from sickchill.oldbeard.common import Quality
from sickchill.show.Show import Show
from sickchill.tv import TVShow


class ShowTests(unittest.TestCase):
    """
    Test shows
    """

    def test_find(self):
        """
        Test find tv shows by indexer_id
        """
        settings.QUALITY_DEFAULT = Quality.FULLHDTV

        settings.showList = []

        show123 = TestTVShow(0, 123)
        show456 = TestTVShow(0, 456)
        show789 = TestTVShow(0, 789)
        shows = [show123, show456, show789]
        shows_duplicate = shows + shows

        test_cases = {
            (False, None): None,
            (False, ""): None,
            (False, "123"): None,
            (False, 123): None,
            (False, 12.3): None,
            (True, None): None,
            (True, ""): None,
            (True, "123"): show123,
            (True, 123): show123,
            (True, 12.3): None,
            (True, 456): show456,
            (True, 789): show789,
        }

        unicode_test_cases = {
            (False, ""): None,
            (False, "123"): None,
            (True, ""): None,
            (True, "123"): show123,
        }

        for tests in test_cases, unicode_test_cases:
            for ((use_shows, indexer_id), result) in tests.items():
                if use_shows:
                    assert Show.find(shows, indexer_id) == result
                else:
                    assert Show.find(None, indexer_id) == result

        with self.assertRaises(MultipleShowObjectsException):
            Show.find(shows_duplicate, 456)

    def test_validate_indexer_id(self):
        """
        Tests if the indexer_id is valid and if so if it returns the right show
        """
        settings.QUALITY_DEFAULT = Quality.FULLHDTV

        settings.showList = []

        show123 = TestTVShow(0, 123)
        show456 = TestTVShow(0, 456)
        show789 = TestTVShow(0, 789)
        settings.showList = [
            show123,
            show456,
            show789,
        ]

        invalid_show_id = ("Invalid show ID", None)

        indexer_id_list = [None, "", "", "123", "123", "456", "456", "789", "789", 123, 456, 789, ["123", "456"], ["123", "456"], [123, 456]]
        results_list = [
            invalid_show_id,
            invalid_show_id,
            invalid_show_id,
            (None, show123),
            (None, show123),
            (None, show456),
            (None, show456),
            (None, show789),
            (None, show789),
            (None, show123),
            (None, show456),
            (None, show789),
            invalid_show_id,
            invalid_show_id,
            invalid_show_id,
        ]

        assert len(indexer_id_list) == len(results_list), "Number of parameters ({0:d}) and results ({1:d}) does not match".format(
            len(indexer_id_list), len(results_list))

        for (index, indexer_id) in enumerate(indexer_id_list):
            assert Show._validate_indexer_id(indexer_id) == results_list[index]


class TestTVShow(TVShow):
    """
    A test `TVShow` object that does not need DB access.
    """

    __test__ = False

    def __init__(self, indexer, indexer_id):
        super().__init__(indexer, indexer_id)

    def loadFromDB(self):
        """
        Override TVShow.loadFromDB to avoid DB access during testing
        """
        pass


if __name__ == "__main__":
    print("=====> Testing {0}".format(__file__))

    SUITE = unittest.TestLoader().loadTestsFromTestCase(ShowTests)
    unittest.TextTestRunner(verbosity=2).run(SUITE)



import logging
import os
import sys
import unittest

from sickchill.oldbeard import browser


class BrowserTestAll(unittest.TestCase):
    """
    Test methods in oldbeard.browser
    """

    def setUp(self):
        self.here = os.path.normpath(os.path.dirname(__file__))

    @unittest.skipUnless(os.name == "nt", "Test on Windows only")
    def test_get_win_drives(self):
        """
        Test getWinDrives
        """
        drives = browser.getWinDrives()
        self.assertIsNotNone(drives)
        self.assertIn("C", drives)

    def test_get_file_list(self):
        """
        Test getFileList
        """
        file_list = browser.getFileList(self.here, True, ["py", "images"])
        self.assertIsNotNone(file_list)
        for entry in file_list:
            assert "name" in entry
            assert "path" in entry
            assert "isImage" in entry
            assert "isFile" in entry
            assert "isAllowed" in entry

            if entry["name"].endswith((".jpg", ".jpeg", ".png", ".tiff", ".gif")):
                assert entry["isImage"]
            else:
                assert not entry["isImage"]

            if entry["name"].endswith(".py") or entry["isImage"]:
                assert entry["isFile"]
            else:
                assert not entry["isFile"]
            assert entry["isAllowed"]

        # folders only
        file_list = browser.getFileList(self.here, False, [])
        self.assertIsNotNone(file_list)
        for entry in file_list:
            assert "name" in entry
            assert "path" in entry
            assert "isImage" in entry
            assert "isFile" in entry
            assert "isAllowed" in entry

            assert not entry["isImage"]
            assert not entry["isFile"]
            assert entry["isAllowed"]

    def test_folders_at_path(self):
        """
        Test foldersAtPath
        """
        test_list = browser.foldersAtPath(os.path.join(self.here, "not_a_real_path"))
        assert test_list[0]["currentPath"] == self.here

        test_list = browser.foldersAtPath("")
        if os.name == "nt":
            assert test_list[0]["currentPath"] == "Root"
            drives = browser.getWinDrives()
            assert len(drives) == len(test_list[1:])
            for item in test_list[1:]:
                assert item["path"].strip(":\\") in drives
        else:
            assert test_list[0]["currentPath"] == "/"

        test_list = browser.foldersAtPath(os.path.join(self.here), includeParent=True)
        assert test_list[0]["currentPath"] == self.here
        assert test_list[1]["name"] == ".."
        assert test_list[1]["path"] == os.path.dirname(self.here)


if __name__ == "__main__":
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger(__name__).setLevel(logging.DEBUG)

    SUITE = unittest.TestLoader().loadTestsFromTestCase(BrowserTestAll)
    unittest.TextTestRunner(verbosity=2).run(SUITE)



import sys
import unittest

from sickchill import settings
from sickchill.oldbeard import common, db, name_cache, scene_exceptions, show_name_helpers
from sickchill.tv import TVShow as Show
from tests import conftest


class SceneTests(conftest.SickChillTestDBCase):
    """
    Test Scene
    """

    def _test_all_possible_show_names(self, name, indexerid=0, expected=None):
        """
        Test all possible show names

        :param name:
        :param indexerid:
        :param expected:
        :return:
        """
        expected = expected or []
        show = Show(1, indexerid)
        show.name = name

        result = show_name_helpers.allPossibleShowNames(show)
        assert len(set(expected).intersection(set(result))) == len(expected)

    def test_all_possible_show_names(self):
        """
        Test all possible show names
        """
        # common.sceneExceptions[-1] = ['Exception Test']
        test_cache_db_con = db.DBConnection("cache.db")
        test_cache_db_con.action("INSERT INTO scene_exceptions (indexer_id, show_name, season) VALUES (?,?,?)", [-1, "Exception Test", -1])
        common.countryList["Full Country Name"] = "FCN"

        self._test_all_possible_show_names("Show Name", expected=["Show Name"])
        self._test_all_possible_show_names("Show Name", -1, expected=["Show Name", "Exception Test"])
        self._test_all_possible_show_names("Show Name FCN", expected=["Show Name FCN", "Show Name (Full Country Name)"])
        self._test_all_possible_show_names("Show Name (FCN)", expected=["Show Name (FCN)", "Show Name (Full Country Name)"])
        self._test_all_possible_show_names("Show Name Full Country Name", expected=["Show Name Full Country Name", "Show Name (FCN)"])
        self._test_all_possible_show_names("Show Name (Full Country Name)", expected=["Show Name (Full Country Name)", "Show Name (FCN)"])

    def test_filter_bad_releases(self):
        """
        Test filtering of bad releases
        """
        settings.IGNORE_WORDS = "GermaN"
        settings.REQUIRE_WORDS = "STUFF"
        assert not show_name_helpers.filter_bad_releases("Show.S02.German.Stuff-Grp")
        assert show_name_helpers.filter_bad_releases("Show.S02.Some.Stuff-Core2HD")
        assert not show_name_helpers.filter_bad_releases("Show.S02.Some.German.Stuff-Grp")
        # assert show_name_helpers.filter_bad_releases('German.Show.S02.Some.Stuff-Grp')
        assert not show_name_helpers.filter_bad_releases("Show.S02.This.Is.German")


class SceneExceptionTestCase(conftest.SickChillTestDBCase):
    """
    Test scene exceptions test case
    """

    def setUp(self):
        """
        Set up tests
        """
        super().setUp()
        scene_exceptions.retrieve_exceptions()

    def test_scene_ex_empty(self):
        """
        Test empty scene exception
        """
        assert scene_exceptions.get_scene_exceptions(0) == []

    def test_scene_ex_babylon_5(self):
        """
        Test scene exceptions for Babylon 5
        """
        assert sorted(scene_exceptions.get_scene_exceptions(70726)) == ["Babylon 5", "Babylon5"]

    def test_scene_ex_by_name(self):
        """
        Test scene exceptions by name
        :return:
        """
        assert scene_exceptions.get_scene_exception_by_name("Babylon5") == (70726, -1)
        assert scene_exceptions.get_scene_exception_by_name("babylon 5") == (70726, -1)
        assert scene_exceptions.get_scene_exception_by_name("Carlos 2010") == (164451, -1)

    def test_scene_ex_by_name_empty(self):
        """
        Test scene exceptions by name are empty
        """
        assert scene_exceptions.get_scene_exception_by_name("nothing useful") == (None, None)

    def test_scene_ex_reset_name_cache(self):
        """
        Test scene exceptions reset name cache
        """
        # clear the exceptions
        test_cache_db_con = db.DBConnection("cache.db")
        test_cache_db_con.action("DELETE FROM scene_exceptions")

        # put something in the cache
        name_cache.add_name("Cached Name", 0)

        # updating should not clear the cache this time since our exceptions didn't change
        scene_exceptions.retrieve_exceptions()
        assert name_cache.get_id_from_name("Cached Name") == 0


if __name__ == "__main__":
    if len(sys.argv) > 1:
        SUITE = unittest.TestLoader().loadTestsFromName("scene_helpers_tests.SceneExceptionTestCase.test_" + sys.argv[1])
        unittest.TextTestRunner(verbosity=2).run(SUITE)
    else:
        SUITE = unittest.TestLoader().loadTestsFromTestCase(SceneTests)
        unittest.TextTestRunner(verbosity=2).run(SUITE)

        SUITE = unittest.TestLoader().loadTestsFromTestCase(SceneExceptionTestCase)
        unittest.TextTestRunner(verbosity=2).run(SUITE)



import sys
import logging
import abc

from ..exceptions import CloudflareSolveError

if sys.version_info >= (3, 4):
    ABC = abc.ABC  # noqa
else:
    ABC = abc.ABCMeta('ABC', (), {})

# ------------------------------------------------------------------------------- #

interpreters = {}

# ------------------------------------------------------------------------------- #


class JavaScriptInterpreter(ABC):

    # ------------------------------------------------------------------------------- #

    @abc.abstractmethod
    def __init__(self, name):
        interpreters[name] = self

    # ------------------------------------------------------------------------------- #

    @classmethod
    def dynamicImport(cls, name):
        if name not in interpreters:
            try:
                __import__('{}.{}'.format(cls.__module__, name))
                if not isinstance(interpreters.get(name), JavaScriptInterpreter):
                    raise ImportError('The interpreter was not initialized.')
            except ImportError:
                logging.error('Unable to load {} interpreter'.format(name))
                raise

        return interpreters[name]

    # ------------------------------------------------------------------------------- #

    @abc.abstractmethod
    def eval(self, jsEnv, js):
        pass

    # ------------------------------------------------------------------------------- #

    def solveChallenge(self, body, domain):
        try:
            return '{0:.10f}'.format(float(self.eval(body, domain)))
        except Exception:
            raise CloudflareSolveError(
                'Error trying to solve Cloudflare IUAM Javascript, they may have changed their technique.'
            )



import datetime
import re
from xml.etree import ElementTree

from babelfish import Country

import sickchill
from sickchill.helper.common import dateFormat
from sickchill.oldbeard import helpers

from ... import logger
from . import generic


class KODIMetadata(generic.GenericMetadata):
    """
    Metadata generation class for KODI 12+.

    The following file structure is used:

    show_root/tvshow.nfo                    (show metadata)
    show_root/fanart.jpg                    (fanart)
    show_root/poster.jpg                    (poster)
    show_root/banner.jpg                    (banner)
    show_root/Season ##/filename.ext        (*)
    show_root/Season ##/filename.nfo        (episode metadata)
    show_root/Season ##/filename-thumb.jpg  (episode thumb)
    show_root/season##-poster.jpg           (season posters)
    show_root/season##-banner.jpg           (season banners)
    show_root/season-all-poster.jpg         (season all poster)
    show_root/season-all-banner.jpg         (season all banner)
    """

    def __init__(
        self,
        show_metadata=False,
        episode_metadata=False,
        fanart=False,
        poster=False,
        banner=False,
        episode_thumbnails=False,
        season_posters=False,
        season_banners=False,
        season_all_poster=False,
        season_all_banner=False,
    ):

        super().__init__(
            show_metadata, episode_metadata, fanart, poster, banner, episode_thumbnails, season_posters, season_banners, season_all_poster, season_all_banner
        )

        self.name = "KODI"

        self.poster_name = "poster.jpg"
        self.season_all_poster_name = "season-all-poster.jpg"

        # web-ui metadata template
        self.eg_show_metadata = "tvshow.nfo"
        self.eg_episode_metadata = "Season##\\<i>filename</i>.nfo"
        self.eg_fanart = "fanart.jpg"
        self.eg_poster = "poster.jpg"
        self.eg_banner = "banner.jpg"
        self.eg_episode_thumbnails = "Season##\\<i>filename</i>-thumb.jpg"
        self.eg_season_posters = "season##-poster.jpg"
        self.eg_season_banners = "season##-banner.jpg"
        self.eg_season_all_poster = "season-all-poster.jpg"
        self.eg_season_all_banner = "season-all-banner.jpg"

    @staticmethod
    def _split_info(info_string):
        return {x.strip().title() for x in re.sub(r"[,/]+", "|", info_string).split("|") if x.strip()}

    def _show_data(self, show_obj):
        """
        Creates an elementTree XML structure for an KODI-style tvshow.nfo and
        returns the resulting data object.

        show_obj: a TVShow instance to create the NFO for
        """
        tv_node = ElementTree.Element("tvshow")

        myShow = sickchill.indexer.series(show_obj)
        if not myShow:
            logger.info("Unable to find show with id {} on {}, skipping it".format(show_obj.indexerid, show_obj.idxr.name))
            return False

        # check for title and id
        if not (getattr(myShow, "seriesName", None) and getattr(myShow, "id", None)):
            logger.info("Incomplete info for show with id {} on {}, skipping it".format(show_obj.indexerid, show_obj.idxr.name))
            return False

        title = ElementTree.SubElement(tv_node, "title")
        title.text = myShow.seriesName

        if getattr(myShow, "rating", None):
            mpaa = ElementTree.SubElement(tv_node, "mpaa")
            mpaa.text = str(myShow.rating)

        if getattr(myShow, "siteRating", None):
            rating = ElementTree.SubElement(tv_node, "rating")
            rating.text = str(myShow.siteRating)

        if getattr(myShow, "firstAired", None):
            try:
                year_text = str(datetime.datetime.strptime(myShow.firstAired, dateFormat).year)
                if year_text:
                    year = ElementTree.SubElement(tv_node, "year")
                    year.text = year_text
            except Exception:
                pass

        if getattr(myShow, "overview", None):
            plot = ElementTree.SubElement(tv_node, "plot")
            plot.text = myShow.overview

        if getattr(myShow, "id", None):
            episodeguide = ElementTree.SubElement(tv_node, "episodeguide")
            episodeguideurl = ElementTree.SubElement(episodeguide, "url")
            episodeguideurl.attrib = {"post": "yes", "cache": "auth.json"}
            episodeguideurl.text = show_obj.idxr.episode_guide_url(show_obj)

            indexerid = ElementTree.SubElement(tv_node, "id")
            indexerid.text = str(myShow.id)

            indexerid = ElementTree.SubElement(tv_node, "tvdbid")
            indexerid.text = str(myShow.id)

        if getattr(myShow, "genre", None) and isinstance(myShow.genre, list):
            for genre in myShow.genre:
                cur_genre = ElementTree.SubElement(tv_node, "genre")
                cur_genre.text = genre

        if show_obj.imdb_info.get("country_codes"):
            for country in self._split_info(show_obj.imdb_info["country_codes"]):
                try:
                    cur_country_name = Country(country.upper()).name.title()
                except Exception:
                    continue

                cur_country = ElementTree.SubElement(tv_node, "country")
                cur_country.text = cur_country_name

        if getattr(myShow, "firstAired", None):
            premiered = ElementTree.SubElement(tv_node, "premiered")
            premiered.text = myShow.firstAired

        if getattr(myShow, "network", None):
            studio = ElementTree.SubElement(tv_node, "studio")
            studio.text = myShow.network.strip()

        data = show_obj.idxr.actors(myShow)
        if data:
            for actor in data:
                cur_actor = ElementTree.SubElement(tv_node, "actor")

                if "name" in actor and actor["name"].strip():
                    cur_actor_name = ElementTree.SubElement(cur_actor, "name")
                    cur_actor_name.text = actor["name"].strip()
                else:
                    continue

                if "role" in actor and actor["role"].strip():
                    cur_actor_role = ElementTree.SubElement(cur_actor, "role")
                    cur_actor_role.text = actor["role"].strip()

                if "image" in actor and actor["image"].strip():
                    cur_actor_thumb = ElementTree.SubElement(cur_actor, "thumb")
                    cur_actor_thumb.text = show_obj.idxr.complete_image_url(actor["image"])

        # Make it purdy
        helpers.indentXML(tv_node)

        data = ElementTree.ElementTree(tv_node)

        return data

    def _ep_data(self, ep_obj):
        """
        Creates an elementTree XML structure for an KODI-style episode.nfo and
        returns the resulting data object.
            show_obj: a TVEpisode instance to create the NFO for
        """

        eps_to_write = [ep_obj] + ep_obj.relatedEps

        myShow = ep_obj.idxr.series_from_episode(ep_obj)
        if not myShow:
            logger.info("Unable to find {} on {} while creating meta files, skipping".format(ep_obj.show.indexerid, ep_obj.idxr.name))
            return

        if len(eps_to_write) > 1:
            rootNode = ElementTree.Element("kodimultiepisode")
        else:
            rootNode = ElementTree.Element("episodedetails")

        # write an NFO containing info for all matching episodes
        for curEpToWrite in eps_to_write:

            myEp = curEpToWrite.idxr.episode(curEpToWrite)
            if not myEp:
                logger.info(
                    "Metadata writer is unable to find episode {0:d}x{1:d} of {2} on {3}..."
                    "has it been removed? Should I delete from db?".format(curEpToWrite.season, curEpToWrite.episode, curEpToWrite.show.name, ep_obj.idxr.name)
                )

                return None

            if ep_obj.airdate != datetime.date.min and not myEp.get("firstAired"):
                myEp["firstAired"] = str(ep_obj.airdate)

            if not myEp.get("episodeName"):
                logger.debug("Not generating nfo because the ep has no title")
                return None

            if not myEp.get("firstAired"):
                logger.debug("Not generating nfo because the ep has no airdate")
                return None

            logger.debug("Creating metadata for episode " + str(ep_obj.season) + "x" + str(ep_obj.episode))

            if len(eps_to_write) > 1:
                episode = ElementTree.SubElement(rootNode, "episodedetails")
            else:
                episode = rootNode

            if myEp.get("episodeName"):
                title = ElementTree.SubElement(episode, "title")
                title.text = myEp["episodeName"]

            if getattr(myShow, "seriesName", None):
                showtitle = ElementTree.SubElement(episode, "showtitle")
                showtitle.text = myShow.seriesName

            season = ElementTree.SubElement(episode, "season")
            season.text = str(curEpToWrite.season)

            episodenum = ElementTree.SubElement(episode, "episode")
            episodenum.text = str(curEpToWrite.episode)

            if myEp.get("id"):
                uniqueid = ElementTree.SubElement(episode, "uniqueid")
                uniqueid.text = str(myEp["id"])

            if curEpToWrite.airdate != datetime.date.min:
                aired = ElementTree.SubElement(episode, "aired")
                aired.text = str(curEpToWrite.airdate)

            if myEp.get("overview"):
                plot = ElementTree.SubElement(episode, "plot")
                plot.text = myEp["overview"]

            if curEpToWrite.season and getattr(myShow, "runtime", None):
                runtime = ElementTree.SubElement(episode, "runtime")
                runtime.text = myShow.runtime

            if myEp.get("airedSeason"):
                displayseason = ElementTree.SubElement(episode, "displayseason")
                displayseason.text = str(myEp["airedSeason"])

            if myEp.get("airedEpisodeNumber"):
                displayepisode = ElementTree.SubElement(episode, "displayepisode")
                displayepisode.text = str(myEp["airedEpisodeNumber"])

            if myEp.get("filename"):
                thumb = ElementTree.SubElement(episode, "thumb")
                thumb.text = ep_obj.idxr.complete_image_url(myEp["filename"])

            if myEp.get("rating") is not None:
                rating = ElementTree.SubElement(episode, "rating")
                rating.text = str(myEp["rating"])

            if myEp.get("writers") and isinstance(myEp["writers"], list):
                for writer in myEp["writers"]:
                    cur_writer = ElementTree.SubElement(episode, "credits")
                    cur_writer.text = writer

            if myEp.get("directors") and isinstance(myEp["directors"], list):
                for director in myEp["directors"]:
                    cur_director = ElementTree.SubElement(episode, "director")
                    cur_director.text = director

            if myEp.get("guestStars") and isinstance(myEp["guestStars"], list):
                for actor in myEp["guestStars"]:
                    cur_actor = ElementTree.SubElement(episode, "actor")
                    cur_actor_name = ElementTree.SubElement(cur_actor, "name")
                    cur_actor_name.text = actor

            for actor in ep_obj.idxr.actors(myShow):
                cur_actor = ElementTree.SubElement(episode, "actor")

                if "name" in actor and actor["name"].strip():
                    cur_actor_name = ElementTree.SubElement(cur_actor, "name")
                    cur_actor_name.text = actor["name"].strip()
                else:
                    continue

                if "role" in actor and actor["role"].strip():
                    cur_actor_role = ElementTree.SubElement(cur_actor, "role")
                    cur_actor_role.text = actor["role"].strip()

                if "image" in actor and actor["image"].strip():
                    cur_actor_thumb = ElementTree.SubElement(cur_actor, "thumb")
                    cur_actor_thumb.text = ep_obj.idxr.complete_image_url(actor["image"])

        # Make it purdy
        helpers.indentXML(rootNode)

        return ElementTree.ElementTree(rootNode)


# present a standard "interface" from the module
metadata_class = KODIMetadata



# -*- coding: utf-8 -*-
import logging
from datetime import datetime
import hashlib
import os
import re
import socket
import struct

import requests
from requests.exceptions import SSLError
from six.moves.xmlrpc_client import ProtocolError

from .exceptions import ServiceUnavailable


logger = logging.getLogger(__name__)


def hash_opensubtitles(video_path):
    """Compute a hash using OpenSubtitles' algorithm.

    :param str video_path: path of the video.
    :return: the hash.
    :rtype: str

    """
    bytesize = struct.calcsize(b'<q')
    with open(video_path, 'rb') as f:
        filesize = os.path.getsize(video_path)
        filehash = filesize
        if filesize < 65536 * 2:
            return
        for _ in range(65536 // bytesize):
            filebuffer = f.read(bytesize)
            (l_value,) = struct.unpack(b'<q', filebuffer)
            filehash += l_value
            filehash &= 0xFFFFFFFFFFFFFFFF  # to remain as 64bit number
        f.seek(max(0, filesize - 65536), 0)
        for _ in range(65536 // bytesize):
            filebuffer = f.read(bytesize)
            (l_value,) = struct.unpack(b'<q', filebuffer)
            filehash += l_value
            filehash &= 0xFFFFFFFFFFFFFFFF
    returnedhash = '%016x' % filehash

    return returnedhash


def hash_thesubdb(video_path):
    """Compute a hash using TheSubDB's algorithm.

    :param str video_path: path of the video.
    :return: the hash.
    :rtype: str

    """
    readsize = 64 * 1024
    if os.path.getsize(video_path) < readsize:
        return
    with open(video_path, 'rb') as f:
        data = f.read(readsize)
        f.seek(-readsize, os.SEEK_END)
        data += f.read(readsize)

    return hashlib.md5(data).hexdigest()


def hash_napiprojekt(video_path):
    """Compute a hash using NapiProjekt's algorithm.

    :param str video_path: path of the video.
    :return: the hash.
    :rtype: str

    """
    readsize = 1024 * 1024 * 10
    with open(video_path, 'rb') as f:
        data = f.read(readsize)
    return hashlib.md5(data).hexdigest()


def hash_shooter(video_path):
    """Compute a hash using Shooter's algorithm

    :param string video_path: path of the video
    :return: the hash
    :rtype: string

    """
    filesize = os.path.getsize(video_path)
    readsize = 4096
    if os.path.getsize(video_path) < readsize * 2:
        return None
    offsets = (readsize, filesize // 3 * 2, filesize // 3, filesize - readsize * 2)
    filehash = []
    with open(video_path, 'rb') as f:
        for offset in offsets:
            f.seek(offset)
            filehash.append(hashlib.md5(f.read(readsize)).hexdigest())
    return ';'.join(filehash)


def sanitize(string, ignore_characters=None):
    """Sanitize a string to strip special characters.

    :param str string: the string to sanitize.
    :param set ignore_characters: characters to ignore.
    :return: the sanitized string.
    :rtype: str

    """
    # only deal with strings
    if string is None:
        return

    ignore_characters = ignore_characters or set()

    # replace some characters with one space
    characters = {'-', ':', '(', ')', '.', ','} - ignore_characters
    if characters:
        string = re.sub(r'[%s]' % re.escape(''.join(characters)), ' ', string)

    # remove some characters
    characters = {'\''} - ignore_characters
    if characters:
        string = re.sub(r'[%s]' % re.escape(''.join(characters)), '', string)

    # replace multiple spaces with one
    string = re.sub(r'\s+', ' ', string)

    # strip and lower case
    return string.strip().lower()


def sanitize_release_group(string):
    """Sanitize a `release_group` string to remove content in square brackets.

    :param str string: the release group to sanitize.
    :return: the sanitized release group.
    :rtype: str

    """
    # only deal with strings
    if string is None:
        return

    # remove content in square brackets
    string = re.sub(r'\[\w+\]', '', string)

    # strip and upper case
    return string.strip().upper()


def timestamp(date):
    """Get the timestamp of the `date`, python2/3 compatible

    :param datetime.datetime date: the utc date.
    :return: the timestamp of the date.
    :rtype: float

    """
    return (date - datetime(1970, 1, 1)).total_seconds()


def matches_title(actual, title, alternative_titles):
    """Whether `actual` matches the `title` or `alternative_titles`

    :param str actual: the actual title to check
    :param str title: the expected title
    :param list alternative_titles: the expected alternative_titles
    :return: whether the actual title matches the title or alternative_titles.
    :rtype: bool

    """
    actual = sanitize(actual)
    title = sanitize(title)
    if actual == title:
        return True

    alternative_titles = set(sanitize(t) for t in alternative_titles)
    if actual in alternative_titles:
        return True

    return actual.startswith(title) and actual[len(title):].strip() in alternative_titles


def handle_exception(e, msg):
    """Handle exception, logging the proper error message followed by `msg`.

    Exception traceback is only logged for specific cases.

    :param exception e: The exception to handle.
    :param str msg: The message to log.
    """
    if isinstance(e, (requests.Timeout, socket.timeout)):
        logger.error('Request timed out. %s', msg)
    elif isinstance(e, (ServiceUnavailable, ProtocolError)):
        # OpenSubtitles raises xmlrpclib.ProtocolError when unavailable
        logger.error('Service unavailable. %s', msg)
    elif isinstance(e, requests.exceptions.HTTPError):
        logger.error('HTTP error %r. %s', e.response.status_code, msg,
                     exc_info=e.response.status_code not in range(500, 600))
    elif isinstance(e, SSLError):
        logger.error('SSL error %r. %s', e.args[0], msg,
                     exc_info=e.args[0] != 'The read operation timed out')
    else:
        logger.exception('Unexpected error. %s', msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2017-8-30

@author: generated by @lolobosse script
'''
LOCALE = [
    ["μόλις τώρα", "σε λίγο"],
    ["%s δευτερόλεπτα πριν", "σε %s δευτερόλεπτα"],
    ["1 λεπτό πριν", "σε 1 λεπτό"],
    ["%s λεπτά πριν", "σε %s λεπτά"],
    ["1 ώρα πριν", "σε 1 ώρα"],
    ["%s ώρες πριν", "σε %s ώρες"],
    ["1 μέρα πριν", "σε 1 μέρα"],
    ["%s μέρες πριν", "σε %s μέρες"],
    ["1 εβδομάδα πριν", "σε 1 εβδομάδα"],
    ["%s εβδομάδες πριν", "σε %s εβδομάδες"],
    ["1 μήνα πριν", "σε 1 μήνα"],
    ["%s μήνες πριν", "σε %s μήνες"],
    ["1 χρόνο πριν", "σε 1 χρόνο"],
    ["%s χρόνια πριν", "σε %s χρόνια"]
]
        


# -*- coding: utf-8 -*-
"""
Subliminal uses `click <http://click.pocoo.org>`_ to provide a powerful :abbr:`CLI (command-line interface)`.

"""
from __future__ import division
from collections import defaultdict
from datetime import timedelta
import glob
import json
import logging
import os
import re

from appdirs import AppDirs
from babelfish import Error as BabelfishError, Language
import click
from dogpile.cache.backends.file import AbstractFileLock
from dogpile.util.readwrite_lock import ReadWriteMutex
from six.moves import configparser

from subliminal import (AsyncProviderPool, Episode, Movie, Video, __version__, check_video, compute_score, get_scores,
                        provider_manager, refine, refiner_manager, region, save_subtitles, scan_video, scan_videos)
from subliminal.core import ARCHIVE_EXTENSIONS, search_external_subtitles

logger = logging.getLogger(__name__)


class MutexLock(AbstractFileLock):
    """:class:`MutexLock` is a thread-based rw lock based on :class:`dogpile.core.ReadWriteMutex`."""
    def __init__(self, filename):
        self.mutex = ReadWriteMutex()

    def acquire_read_lock(self, wait):
        ret = self.mutex.acquire_read_lock(wait)
        return wait or ret

    def acquire_write_lock(self, wait):
        ret = self.mutex.acquire_write_lock(wait)
        return wait or ret

    def release_read_lock(self):
        return self.mutex.release_read_lock()

    def release_write_lock(self):
        return self.mutex.release_write_lock()


class Config(object):
    """A :class:`~configparser.ConfigParser` wrapper to store configuration.

    Interaction with the configuration is done with the properties.

    :param str path: path to the configuration file.

    """
    def __init__(self, path):
        #: Path to the configuration file
        self.path = path

        #: The underlying configuration object
        self.config = configparser.SafeConfigParser()
        self.config.add_section('general')
        self.config.set('general', 'languages', json.dumps(['en']))
        self.config.set('general', 'providers', json.dumps(sorted([p.name for p in provider_manager])))
        self.config.set('general', 'refiners', json.dumps(sorted([r.name for r in refiner_manager])))
        self.config.set('general', 'single', str(0))
        self.config.set('general', 'embedded_subtitles', str(1))
        self.config.set('general', 'age', str(int(timedelta(weeks=2).total_seconds())))
        self.config.set('general', 'hearing_impaired', str(1))
        self.config.set('general', 'min_score', str(0))

    def read(self):
        """Read the configuration from :attr:`path`"""
        self.config.read(self.path)

    def write(self):
        """Write the configuration to :attr:`path`"""
        with open(self.path, 'w') as f:
            self.config.write(f)

    @property
    def languages(self):
        return {Language.fromietf(l) for l in json.loads(self.config.get('general', 'languages'))}

    @languages.setter
    def languages(self, value):
        self.config.set('general', 'languages', json.dumps(sorted([str(l) for l in value])))

    @property
    def providers(self):
        return json.loads(self.config.get('general', 'providers'))

    @providers.setter
    def providers(self, value):
        self.config.set('general', 'providers', json.dumps(sorted([p.lower() for p in value])))

    @property
    def refiners(self):
        return json.loads(self.config.get('general', 'refiners'))

    @refiners.setter
    def refiners(self, value):
        self.config.set('general', 'refiners', json.dumps([r.lower() for r in value]))

    @property
    def single(self):
        return self.config.getboolean('general', 'single')

    @single.setter
    def single(self, value):
        self.config.set('general', 'single', str(int(value)))

    @property
    def embedded_subtitles(self):
        return self.config.getboolean('general', 'embedded_subtitles')

    @embedded_subtitles.setter
    def embedded_subtitles(self, value):
        self.config.set('general', 'embedded_subtitles', str(int(value)))

    @property
    def age(self):
        return timedelta(seconds=self.config.getint('general', 'age'))

    @age.setter
    def age(self, value):
        self.config.set('general', 'age', str(int(value.total_seconds())))

    @property
    def hearing_impaired(self):
        return self.config.getboolean('general', 'hearing_impaired')

    @hearing_impaired.setter
    def hearing_impaired(self, value):
        self.config.set('general', 'hearing_impaired', str(int(value)))

    @property
    def min_score(self):
        return self.config.getfloat('general', 'min_score')

    @min_score.setter
    def min_score(self, value):
        self.config.set('general', 'min_score', str(value))

    @property
    def provider_configs(self):
        rv = {}
        for provider in provider_manager:
            if self.config.has_section(provider.name):
                rv[provider.name] = {k: v for k, v in self.config.items(provider.name)}
        return rv

    @provider_configs.setter
    def provider_configs(self, value):
        # loop over provider configurations
        for provider, config in value.items():
            # create the corresponding section if necessary
            if not self.config.has_section(provider):
                self.config.add_section(provider)

            # add config options
            for k, v in config.items():
                self.config.set(provider, k, v)

    @property
    def refiner_configs(self):
        rv = {}
        for refiner in refiner_manager:
            if self.config.has_section(refiner.name):
                rv[refiner.name] = {k: v for k, v in self.config.items(refiner.name)}
        return rv

    @refiner_configs.setter
    def refiner_configs(self, value):
        # loop over refiner configurations
        for refiner, config in value.items():
            # create the corresponding section if necessary
            if not self.config.has_section(refiner):
                self.config.add_section(refiner)

            # add config options
            for k, v in config.items():
                self.config.set(refiner, k, v)


class LanguageParamType(click.ParamType):
    """:class:`~click.ParamType` for languages that returns a :class:`~babelfish.language.Language`"""
    name = 'language'

    def convert(self, value, param, ctx):
        try:
            return Language.fromietf(value)
        except BabelfishError:
            self.fail('%s is not a valid language' % value)


LANGUAGE = LanguageParamType()


class AgeParamType(click.ParamType):
    """:class:`~click.ParamType` for age strings that returns a :class:`~datetime.timedelta`

    An age string is in the form `number + identifier` with possible identifiers:

        * ``w`` for weeks
        * ``d`` for days
        * ``h`` for hours

    The form can be specified multiple times but only with that idenfier ordering. For example:

        * ``1w2d4h`` for 1 week, 2 days and 4 hours
        * ``2w`` for 2 weeks
        * ``3w6h`` for 3 weeks and 6 hours

    """
    name = 'age'

    def convert(self, value, param, ctx):
        match = re.match(r'^(?:(?P<weeks>\d+?)w)?(?:(?P<days>\d+?)d)?(?:(?P<hours>\d+?)h)?$', value)
        if not match:
            self.fail('%s is not a valid age' % value)

        return timedelta(**{k: int(v) for k, v in match.groupdict(0).items()})


AGE = AgeParamType()

PROVIDER = click.Choice(sorted(provider_manager.names()))

REFINER = click.Choice(sorted(refiner_manager.names()))

dirs = AppDirs('subliminal')
cache_file = 'subliminal.dbm'
config_file = 'config.ini'


@click.group(context_settings={'max_content_width': 100}, epilog='Suggestions and bug reports are greatly appreciated: '
             'https://github.com/Diaoul/subliminal/')
@click.option('--addic7ed', type=click.STRING, nargs=2, metavar='USERNAME PASSWORD', help='Addic7ed configuration.')
@click.option('--legendastv', type=click.STRING, nargs=2, metavar='USERNAME PASSWORD', help='LegendasTV configuration.')
@click.option('--opensubtitles', type=click.STRING, nargs=2, metavar='USERNAME PASSWORD',
              help='OpenSubtitles configuration.')
@click.option('--omdb', type=click.STRING, nargs=1, metavar='APIKEY', help='OMDB API key.')
@click.option('--cache-dir', type=click.Path(writable=True, file_okay=False), default=dirs.user_cache_dir,
              show_default=True, expose_value=True, help='Path to the cache directory.')
@click.option('--debug', is_flag=True, help='Print useful information for debugging subliminal and for reporting bugs.')
@click.version_option(__version__)
@click.pass_context
def subliminal(ctx, addic7ed, legendastv, opensubtitles, omdb, cache_dir, debug):
    """Subtitles, faster than your thoughts."""
    # create cache directory
    try:
        os.makedirs(cache_dir)
    except OSError:
        if not os.path.isdir(cache_dir):
            raise

    # configure cache
    region.configure('dogpile.cache.dbm', expiration_time=timedelta(days=30),
                     arguments={'filename': os.path.join(cache_dir, cache_file), 'lock_factory': MutexLock})

    # configure logging
    if debug:
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(logging.BASIC_FORMAT))
        logging.getLogger('subliminal').addHandler(handler)
        logging.getLogger('subliminal').setLevel(logging.DEBUG)

    ctx.obj = {
        'provider_configs': {},
        'refiner_configs': {}
    }

    # provider configs
    if addic7ed:
        ctx.obj['provider_configs']['addic7ed'] = {'username': addic7ed[0], 'password': addic7ed[1]}
    if legendastv:
        ctx.obj['provider_configs']['legendastv'] = {'username': legendastv[0], 'password': legendastv[1]}
    if opensubtitles:
        ctx.obj['provider_configs']['opensubtitles'] = {'username': opensubtitles[0], 'password': opensubtitles[1]}
        ctx.obj['provider_configs']['opensubtitlesvip'] = {'username': opensubtitles[0], 'password': opensubtitles[1]}

    # refiner configs
    if omdb:
        ctx.obj['refiner_configs']['omdb'] = {'apikey': omdb}


@subliminal.command()
@click.option('--clear-subliminal', is_flag=True, help='Clear subliminal\'s cache. Use this ONLY if your cache is '
              'corrupted or if you experience issues.')
@click.pass_context
def cache(ctx, clear_subliminal):
    """Cache management."""
    if clear_subliminal:
        for file in glob.glob(os.path.join(ctx.parent.params['cache_dir'], cache_file) + '*'):
            os.remove(file)
        click.echo('Subliminal\'s cache cleared.')
    else:
        click.echo('Nothing done.')


@subliminal.command()
@click.option('-l', '--language', type=LANGUAGE, required=True, multiple=True, help='Language as IETF code, '
              'e.g. en, pt-BR (can be used multiple times).')
@click.option('-p', '--provider', type=PROVIDER, multiple=True, help='Provider to use (can be used multiple times).')
@click.option('-r', '--refiner', type=REFINER, multiple=True, help='Refiner to use (can be used multiple times).')
@click.option('-a', '--age', type=AGE, help='Filter videos newer than AGE, e.g. 12h, 1w2d.')
@click.option('-d', '--directory', type=click.STRING, metavar='DIR', help='Directory where to save subtitles, '
              'default is next to the video file.')
@click.option('-e', '--encoding', type=click.STRING, metavar='ENC', help='Subtitle file encoding, default is to '
              'preserve original encoding.')
@click.option('-s', '--single', is_flag=True, default=False, help='Save subtitle without language code in the file '
              'name, i.e. use .srt extension. Do not use this unless your media player requires it.')
@click.option('-f', '--force', is_flag=True, default=False, help='Force download even if a subtitle already exist.')
@click.option('-hi', '--hearing-impaired', is_flag=True, default=False, help='Prefer hearing impaired subtitles.')
@click.option('-m', '--min-score', type=click.IntRange(0, 100), default=0, help='Minimum score for a subtitle '
              'to be downloaded (0 to 100).')
@click.option('-w', '--max-workers', type=click.IntRange(1, 50), default=None, help='Maximum number of threads to use.')
@click.option('-z/-Z', '--archives/--no-archives', default=True, show_default=True, help='Scan archives for videos '
              '(supported extensions: %s).' % ', '.join(ARCHIVE_EXTENSIONS))
@click.option('-v', '--verbose', count=True, help='Increase verbosity.')
@click.argument('path', type=click.Path(), required=True, nargs=-1)
@click.pass_obj
def download(obj, provider, refiner, language, age, directory, encoding, single, force, hearing_impaired, min_score,
             max_workers, archives, verbose, path):
    """Download best subtitles.

    PATH can be an directory containing videos, a video file path or a video file name. It can be used multiple times.

    If an existing subtitle is detected (external or embedded) in the correct language, the download is skipped for
    the associated video.

    """
    # process parameters
    language = set(language)

    # scan videos
    videos = []
    ignored_videos = []
    errored_paths = []
    with click.progressbar(path, label='Collecting videos', item_show_func=lambda p: p or '') as bar:
        for p in bar:
            logger.debug('Collecting path %s', p)

            # non-existing
            if not os.path.exists(p):
                try:
                    video = Video.fromname(p)
                except:
                    logger.exception('Unexpected error while collecting non-existing path %s', p)
                    errored_paths.append(p)
                    continue
                if not force:
                    video.subtitle_languages |= set(search_external_subtitles(video.name, directory=directory).values())

                if check_video(video, languages=language, age=age, undefined=single):
                    refine(video, episode_refiners=refiner, movie_refiners=refiner,
                           refiner_configs=obj['refiner_configs'],
                           embedded_subtitles=not force, providers=provider, languages=language)
                    videos.append(video)
                continue

            # directories
            if os.path.isdir(p):
                try:
                    scanned_videos = scan_videos(p, age=age, archives=archives)
                except:
                    logger.exception('Unexpected error while collecting directory path %s', p)
                    errored_paths.append(p)
                    continue
                for video in scanned_videos:
                    if not force:
                        video.subtitle_languages |= set(search_external_subtitles(video.name,
                                                                                  directory=directory).values())
                    if check_video(video, languages=language, age=age, undefined=single):
                        refine(video, episode_refiners=refiner, movie_refiners=refiner,
                               refiner_configs=obj['refiner_configs'], embedded_subtitles=not force,
                               providers=provider, languages=language)
                        videos.append(video)
                    else:
                        ignored_videos.append(video)
                continue

            # other inputs
            try:
                video = scan_video(p)
            except:
                logger.exception('Unexpected error while collecting path %s', p)
                errored_paths.append(p)
                continue
            if not force:
                video.subtitle_languages |= set(search_external_subtitles(video.name, directory=directory).values())
            if check_video(video, languages=language, age=age, undefined=single):
                refine(video, episode_refiners=refiner, movie_refiners=refiner,
                       refiner_configs=obj['refiner_configs'], embedded_subtitles=not force,
                       providers=provider, languages=language)
                videos.append(video)
            else:
                ignored_videos.append(video)

    # output errored paths
    if verbose > 0:
        for p in errored_paths:
            click.secho('%s errored' % p, fg='red')

    # output ignored videos
    if verbose > 1:
        for video in ignored_videos:
            click.secho('%s ignored - subtitles: %s / age: %d day%s' % (
                os.path.split(video.name)[1],
                ', '.join(str(s) for s in video.subtitle_languages) or 'none',
                video.age.days,
                's' if video.age.days > 1 else ''
            ), fg='yellow')

    # report collected videos
    click.echo('%s video%s collected / %s video%s ignored / %s error%s' % (
        click.style(str(len(videos)), bold=True, fg='green' if videos else None),
        's' if len(videos) > 1 else '',
        click.style(str(len(ignored_videos)), bold=True, fg='yellow' if ignored_videos else None),
        's' if len(ignored_videos) > 1 else '',
        click.style(str(len(errored_paths)), bold=True, fg='red' if errored_paths else None),
        's' if len(errored_paths) > 1 else '',
    ))

    # exit if no video collected
    if not videos:
        return

    # download best subtitles
    downloaded_subtitles = defaultdict(list)
    with AsyncProviderPool(max_workers=max_workers, providers=provider, provider_configs=obj['provider_configs']) as p:
        with click.progressbar(videos, label='Downloading subtitles',
                               item_show_func=lambda v: os.path.split(v.name)[1] if v is not None else '') as bar:
            for v in bar:
                scores = get_scores(v)
                subtitles = p.download_best_subtitles(p.list_subtitles(v, language - v.subtitle_languages),
                                                      v, language, min_score=scores['hash'] * min_score / 100,
                                                      hearing_impaired=hearing_impaired, only_one=single)
                downloaded_subtitles[v] = subtitles

        if p.discarded_providers:
            click.secho('Some providers have been discarded due to unexpected errors: %s' %
                        ', '.join(p.discarded_providers), fg='yellow')

    # save subtitles
    total_subtitles = 0
    for v, subtitles in downloaded_subtitles.items():
        saved_subtitles = save_subtitles(v, subtitles, single=single, directory=directory, encoding=encoding)
        total_subtitles += len(saved_subtitles)

        if verbose > 0:
            click.echo('%s subtitle%s downloaded for %s' % (click.style(str(len(saved_subtitles)), bold=True),
                                                            's' if len(saved_subtitles) > 1 else '',
                                                            os.path.split(v.name)[1]))

        if verbose > 1:
            for s in saved_subtitles:
                matches = s.get_matches(v)
                score = compute_score(s, v)

                # score color
                score_color = None
                scores = get_scores(v)
                if isinstance(v, Movie):
                    if score < scores['title']:
                        score_color = 'red'
                    elif score < scores['title'] + scores['year'] + scores['release_group']:
                        score_color = 'yellow'
                    else:
                        score_color = 'green'
                elif isinstance(v, Episode):
                    if score < scores['series'] + scores['season'] + scores['episode']:
                        score_color = 'red'
                    elif score < scores['series'] + scores['season'] + scores['episode'] + scores['release_group']:
                        score_color = 'yellow'
                    else:
                        score_color = 'green'

                # scale score from 0 to 100 taking out preferences
                scaled_score = score
                if s.hearing_impaired == hearing_impaired:
                    scaled_score -= scores['hearing_impaired']
                scaled_score *= 100 / scores['hash']

                # echo some nice colored output
                click.echo('  - [{score}] {language} subtitle from {provider_name} (match on {matches})'.format(
                    score=click.style('{:5.1f}'.format(scaled_score), fg=score_color, bold=score >= scores['hash']),
                    language=s.language.name if s.language.country is None else '%s (%s)' % (s.language.name,
                                                                                             s.language.country.name),
                    provider_name=s.provider_name,
                    matches=', '.join(sorted(matches, key=scores.get, reverse=True))
                ))

    if verbose == 0:
        click.echo('Downloaded %s subtitle%s' % (click.style(str(total_subtitles), bold=True),
                                                 's' if total_subtitles > 1 else ''))



"""
Test the post processor queue
"""
import datetime
import os.path
import time
import unittest

import sickchill.oldbeard
from sickchill import settings
from sickchill.oldbeard.post_processing_queue import PostProcessorTask, ProcessingQueue
from tests import conftest

CHECK_CLEARS = False


class PostProcessorQueueTests(conftest.SickChillTestPostProcessorCase):
    """
    Test the post processor queue
    """

    def __init__(self, testCaseNames):
        super().__init__(testCaseNames)
        self.queue = None

    def setUp(self):
        super().setUp()
        self.queue = sickchill.oldbeard.scheduler.Scheduler(
            ProcessingQueue(),
            run_delay=datetime.timedelta(seconds=0),
            cycleTime=datetime.timedelta(seconds=1),
            threadName="POSTPROCESSOR",
        )
        self.queue.enable = True
        self.queue.start()

    def tearDown(self):
        self.queue.stop.set()
        super().tearDown()

    def test_post_processor_queue_spam(self):
        settings.TV_DOWNLOAD_DIR = os.path.abspath(".")
        for i in range(100):
            result = self.queue.action.add_item(settings.TV_DOWNLOAD_DIR, method="move", mode=("manual", "auto")[i % 2])
            self.assertIsNotNone(result)
            assert self.queue.action.queue_length()["auto"] <= 1
            assert self.queue.action.queue_length()["manual"] <= 1
            assert len(self.queue.action.queue) <= 2

        for task in self.queue.action.queue:
            self.assertIsInstance(task, PostProcessorTask)

        if CHECK_CLEARS:
            cleared = False
            timeout = 300
            for i in range(timeout):
                if not len(self.queue.action.queue):
                    cleared = True
                    break
                time.sleep(1)

            print(self.queue.action.queue)
            if cleared:
                print("cleared after {}".format(i))
            for task in self.queue.action.queue:
                print(task.last_result)

            assert cleared, "The queue did not empty after {timeout} seconds".format(timeout=timeout)


if __name__ == "__main__":
    print("==================")
    print("STARTING - Post Processor Queue TESTS")
    print("==================")
    print("######################################################################")
    SUITE = unittest.TestLoader().loadTestsFromTestCase(PostProcessorQueueTests)
    unittest.TextTestRunner(verbosity=2).run(SUITE)



# -*- coding: utf-8 -*-
import codecs
import logging
import os

import chardet
import pysrt

from six import text_type

logger = logging.getLogger(__name__)

#: Subtitle extensions
SUBTITLE_EXTENSIONS = ('.srt', '.sub', '.smi', '.txt', '.ssa', '.ass', '.mpl')


class Subtitle(object):
    """Base class for subtitle.

    :param language: language of the subtitle.
    :type language: :class:`~babelfish.language.Language`
    :param bool hearing_impaired: whether or not the subtitle is hearing impaired.
    :param page_link: URL of the web page from which the subtitle can be downloaded.
    :type page_link: str
    :param encoding: Text encoding of the subtitle.
    :type encoding: str

    """
    #: Name of the provider that returns that class of subtitle
    provider_name = ''

    def __init__(self, language, hearing_impaired=False, page_link=None, encoding=None):
        #: Language of the subtitle
        self.language = language

        #: Whether or not the subtitle is hearing impaired
        self.hearing_impaired = hearing_impaired

        #: URL of the web page from which the subtitle can be downloaded
        self.page_link = page_link

        #: Content as bytes
        self.content = None

        #: Encoding to decode with when accessing :attr:`text`
        self.encoding = None

        # validate the encoding
        if encoding:
            try:
                self.encoding = codecs.lookup(encoding).name
            except (TypeError, LookupError):
                logger.debug('Unsupported encoding %s', encoding)

    @property
    def id(self):
        """Unique identifier of the subtitle"""
        raise NotImplementedError

    @property
    def info(self):
        """Info of the subtitle, human readable. Usually the subtitle name for GUI rendering"""
        raise NotImplementedError

    @property
    def text(self):
        """Content as string

        If :attr:`encoding` is None, the encoding is guessed with :meth:`guess_encoding`

        """
        if not self.content:
            return

        if not isinstance(self.content, text_type):
            if self.encoding:
                return self.content.decode(self.encoding, errors='replace')

            guessed_encoding = self.guess_encoding()
            if guessed_encoding:
                return self.content.decode(guessed_encoding, errors='replace')

            return None

        return self.content

    def is_valid(self):
        """Check if a :attr:`text` is a valid SubRip format.

        :return: whether or not the subtitle is valid.
        :rtype: bool

        """
        if not self.text:
            return False

        try:
            pysrt.from_string(self.text, error_handling=pysrt.ERROR_RAISE)
        except pysrt.Error as e:
            if e.args[0] < 80:
                return False

        return True

    def guess_encoding(self):
        """Guess encoding using the language, falling back on chardet.

        :return: the guessed encoding.
        :rtype: str

        """
        logger.info('Guessing encoding for language %s', self.language)

        # always try utf-8 first
        encodings = ['utf-8']

        # add language-specific encodings
        if self.language.alpha3 == 'zho':
            encodings.extend(['gb18030', 'big5'])
        elif self.language.alpha3 == 'jpn':
            encodings.append('shift-jis')
        elif self.language.alpha3 == 'ara':
            encodings.append('windows-1256')
        elif self.language.alpha3 == 'heb':
            encodings.append('windows-1255')
        elif self.language.alpha3 == 'tur':
            encodings.extend(['iso-8859-9', 'windows-1254'])
        elif self.language.alpha3 == 'pol':
            # Eastern European Group 1
            encodings.extend(['windows-1250'])
        elif self.language.alpha3 == 'bul':
            # Eastern European Group 2
            encodings.extend(['windows-1251'])
        else:
            # Western European (windows-1252)
            encodings.append('latin-1')

        # try to decode
        logger.debug('Trying encodings %r', encodings)
        for encoding in encodings:
            try:
                self.content.decode(encoding)
            except UnicodeDecodeError:
                pass
            else:
                logger.info('Guessed encoding %s', encoding)
                return encoding

        logger.warning('Could not guess encoding from language')

        # fallback on chardet
        encoding = chardet.detect(self.content)['encoding']
        logger.info('Chardet found encoding %s', encoding)

        return encoding

    def get_path(self, video, single=False):
        """Get the subtitle path using the `video`, `language` and `extension`.

        :param video: path to the video.
        :type video: :class:`~subliminal.video.Video`
        :param bool single: save a single subtitle, default is to save one subtitle per language.
        :return: path of the subtitle.
        :rtype: str

        """
        return get_subtitle_path(video.name, None if single else self.language)

    def get_matches(self, video):
        """Get the matches against the `video`.

        :param video: the video to get the matches with.
        :type video: :class:`~subliminal.video.Video`
        :return: matches of the subtitle.
        :rtype: set

        """
        raise NotImplementedError

    def __hash__(self):
        return hash(self.provider_name + '-' + self.id)

    def __repr__(self):
        return '<%s %r [%s]>' % (self.__class__.__name__, self.id, self.language)


def get_subtitle_path(video_path, language=None, extension='.srt'):
    """Get the subtitle path using the `video_path` and `language`.

    :param str video_path: path to the video.
    :param language: language of the subtitle to put in the path.
    :type language: :class:`~babelfish.language.Language`
    :param str extension: extension of the subtitle.
    :return: path of the subtitle.
    :rtype: str

    """
    subtitle_root = os.path.splitext(video_path)[0]

    if language:
        subtitle_root += '.' + str(language)

    return subtitle_root + extension


def fix_line_ending(content):
    """Fix line ending of `content` by changing it to \n.

    :param bytes content: content of the subtitle.
    :return: the content with fixed line endings.
    :rtype: bytes

    """
    return content.replace(b'\r\n', b'\n').replace(b'\r', b'\n')
